# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_operating_items_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE 
from RCAIDE.Framework.Core    import Units ,  Data

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Operating Items Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_operating_items_weight(vehicle):
    """ Calculate the weight of operating items, including:
        - crew
        - baggage
        - unusable fuel
        - engine oil
        - passenger service
        - ammunition and non-fixed weapons
        - cargo containers

        Assumptions:
            If no tanks are specified, 5 fuel tanks are assumed (includes main and auxiliary tanks)
            If the number of coach seats is not defined, then it assumed that 5% of
                of the seats are first class and an additional 10 % are business class.
            If the number of coach seats is defined, then the additional seats are 1/4 first class
                and 3/4 business class

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.networks: data dictionary containing all propulsion properties
                    -.number_of_engines: number of engines
                    -.sealevel_static_thrust: thrust at sea level               [N]
                -.reference_area: wing surface area                             [m^2]
                -.mass_properties.max_takeoff: MTOW                             [kilograms]
                -.passengers: number of passengers in aircraft
                -.design_mach_number: design mach number for cruise flight
                -.design_range: design range of aircraft                        [nmi]
                -.mass_properties.cargo: weight of cargo carried                [kilograms]

        Outputs:
            output - data dictionary with weights                               [kilograms]
                    - output.oper_items: unusable fuel, engine oil, passenger service weight and cargo containers
                    - output.flight_crew: flight crew weight
                    - output.flight_attendants: flight attendants weight
                    - output.total: total operating items weight

        Properties Used:
            N/A
    """ 
    NENG =  0 
    for network in  vehicle.networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbojet):
                ref_propulsor = propulsor  
                NENG  += 1   
    
    THRUST          = ref_propulsor.sealevel_static_thrust * 1 / Units.lbf
    SW              = vehicle.reference_area / Units.ft ** 2
    FMXTOT          = vehicle.mass_properties.max_zero_fuel / Units.lbs
    DESRNG          = vehicle.flight_envelope.design_range / Units.nmi
    VMAX            = vehicle.flight_envelope.design_mach_number   
    
    number_of_tanks = 0  
    for network in  vehicle.networks:
        for fuel_line in network.fuel_lines:
            for fuel_tank in fuel_line.fuel_tanks:
                number_of_tanks += 1 
    if number_of_tanks == 0:
        number_of_tanks = 5    
    
    WUF             = 11.5 * NENG * THRUST ** 0.2 + 0.07 * SW + 1.6 * number_of_tanks * FMXTOT ** 0.28  # unusable fuel weight
    WOIL            = 0.082 * NENG * THRUST ** 0.65  # engine oil weight
    
    for fuselage in  vehicle.fuselages: 
        if hasattr(fuselage, 'number_coach_seats'):
            NPT = fuselage.number_coach_seats  # number of economy passengers
            NPF = (vehicle.passengers - NPT) / 4.  # number of first clss passengers
            NPB = vehicle.passengers - NPF - NPT  # number of bussines passengers
        else:
            NPF = vehicle.passengers / 20.
            NPB = vehicle.passengers / 10.
            NPT = vehicle.passengers - NPF - NPB
    vehicle.NPF = NPF
    vehicle.NPB = NPB
    vehicle.NPT = NPT
    WSRV        = (5.164 * NPF + 3.846 * NPB + 2.529 * NPT) * (DESRNG / VMAX) ** 0.255  # passenger service weight
    WCON        = 175 * np.ceil(vehicle.mass_properties.cargo / Units.lbs * 1. / 950)  # cargo container weight

    if vehicle.passengers >= 150:
        NFLCR = 3  # number of flight crew
        NGALC = 1 + np.floor(vehicle.passengers / 250.)  # number of galley crew
    else:
        NFLCR = 2
        NGALC = 0
    if vehicle.passengers < 51:
        NFLA = 1  # number of flight attendants, NSTU in FLOPS
    else:
        NFLA = 1 + np.floor(vehicle.passengers / 40.)

    WFLAAB = NFLA * 155 + NGALC * 200  # flight attendant weight, WSTUAB in FLOPS
    WFLCRB = NFLCR * 225  # flight crew and baggage weight

    output                           = Data()
    output.misc = WUF * Units.lbs + WOIL * Units.lbs + WSRV * Units.lbs + WCON * Units.lbs
    output.flight_crew               = WFLCRB * Units.lbs
    output.flight_attendants         = WFLAAB * Units.lbs
    output.total                     = output.misc + output.flight_crew + \
                                       output.flight_attendants
    return output



# RCAIDE/Library/Missions/Segments/Cruise/Constant_Pitch_Rate_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# Package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant acceleration and constant altitude

    Source:
    N/A

    Inputs:
    segment.altitude                [meters]
    segment.pitch_initial           [radians]
    segment.pitch_final             [radians]
    segment.pitch_rate              [radians/second]

    Outputs:
    conditions.frames.body.inertial_rotations   [radians/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """       
    
    # unpack
    alt        = segment.altitude 
    T0         = segment.pitch_initial
    Tf         = segment.pitch_final 
    theta_dot  = segment.pitch_rate   
    conditions = segment.state.conditions 
    state      = segment.state
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
        
    # check for initial pitch
    if T0 is None:
        T0  =  state.initials.conditions.frames.body.inertial_rotations[-1,1]
        segment.pitch_initial = T0
    
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_final   = (Tf-T0)/theta_dot + t_initial
    t_nondim  = state.numerics.dimensionless.control_points
    time      = t_nondim * (t_final-t_initial) + t_initial
    
    # set the body angle
    body_angle = theta_dot*time + T0
    segment.state.conditions.frames.body.inertial_rotations[:,1] = body_angle[:,0]    
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]
    
    


# RCAIDE/Library/Missions/Segments/Climb/Constant_Speed_Constant_Angle_Noise.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def expand_state(segment):
    
    """Makes all vectors in the state the same size. Determines the minimum amount of points needed to get data for noise certification.

    Assumptions:
    Half second intervals for certification requirements. Fixed microphone position

    Source:
    N/A

    Inputs:
    state.numerics.number_of_control_points  [Unitless]

    Outputs:
    N/A

    Properties Used:
    Position of the flyover microphone is 6500 meters
    """          
    
    # unpack
    climb_angle  = segment.climb_angle
    air_speed    = segment.air_speed   
    conditions   = segment.state.conditions
    
    #Necessary input for determination of noise trajectory    
    dt = 0.5  #time step in seconds for noise calculation - Certification requirement    
    x0 = 6500 #Position of the Flyover microphone relatively to the break-release point
    
    # process velocity vector
    v_x=air_speed*np.cos(climb_angle)
    
    #number of time steps (space discretization)
    total_time=(x0+500)/v_x    
    n_points   = np.int(np.ceil(total_time/dt +1))       
    
    segment.state.numerics.number_of_control_points = n_points
    
    segment.state.expand_rows(n_points,override=True)      
    
    return

# ----------------------------------------------------------------------
#  Initialize Conditions
# ---------------------------------------------------------------------- 
def initialize_conditions(segment):
    """Gets the overall time step for the segment type.
    
    Assumptions:
    Constant true airspeed, with a constant climb angle. This segment is specically created for noise calculations.

    Source:
    N/A

    Inputs:
    segment.climb_angle                         [radians]
    segment.air_speed                           [meter/second]
    segment.altitude_start                      [meters]
    segment.altitude_end                        [meters]
    state.numerics.dimensionless.control_points [Unitless]
    conditions.freestream.density               [kilograms/meter^3]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]

    Properties Used:
    N/A
    """     
    
    dt=0.5  #time step in seconds for noise calculation
    
    # unpack
    climb_angle = segment.climb_angle
    air_speed   = segment.air_speed     
    beta        = segment.sideslip_angle
    t_nondim    = segment.state.numerics.dimensionless.control_points
    conditions  = segment.state.conditions  

    # check for initial velocity
    if air_speed is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # process velocity vector
    v_mag = air_speed
    v_x   = np.cos(beta)*v_mag * np.cos(climb_angle)
    v_y   = np.sin(beta)*v_mag * np.cos(climb_angle)
    v_z   = -v_mag * np.sin(climb_angle)    

    #initial altitude
    alt0 = 10.668   #(35ft)
    altf = alt0 + (-v_z)*dt*len(t_nondim)

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0    
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context



# RCAIDE/Library/Methods/Weights/Correlations/Common/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_landing_gear_weight        import compute_landing_gear_weight    
from .compute_payload_weight             import compute_payload_weight         
from .compute_main_wing_weight           import compute_main_wing_weight       
from .compute_operating_empty_weight     import compute_operating_empty_weight
from .compute_systems_weight             import compute_systems_weight


# RCAIDE/Library/Methods/Emissions/Chemical_Reactor_Network_Method/build_CRN_EI_surrogates.py
#  
# Created: June 2024, M. Clarke, M. Guidotti 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Core import  Data 

# package imports 
from scipy.interpolate  import RegularGridInterpolator

# ----------------------------------------------------------------------------------------------------------------------
#  build_CRN_EI_surrogates
# ---------------------------------------------------------------------------------------------------------------------- 
def build_CRN_EI_surrogates(emissions):

    """
    Builds surrogate models for emission indices using Chemical Reactor Network (CRN) training data.

    Parameters
    ----------
    emissions : Data 
        Container of emission data and surrogate models
        
        - surrogates : Data
            Container for surrogate models
            
        - training : Data
            Training data container
            
            - pressure : ndarray
                Array of pressure values [Pa]
            - temperature : ndarray
                Array of temperature values [K]
            - air_mass_flowrate : ndarray
                Array of air mass flow rates [kg/s]
            - fuel_to_air_ratio : ndarray
                Array of fuel-to-air ratios [-]
            - EI_CO2 : ndarray
                CO2 emission index training data [kg_CO2/kg_fuel]
            - EI_CO : ndarray
                CO emission index training data [kg_CO/kg_fuel]
            - EI_H2O : ndarray
                H2O emission index training data [kg_H2O/kg_fuel]
            - EI_NO : ndarray
                NO emission index training data [kg_NO/kg_fuel]
            - EI_NO2 : ndarray
                NO2 emission index training data [kg_NO2/kg_fuel]

    Returns
    -------
    Updates the emissions.surrogates Data structure with interpolation functions

    Notes
    -----
    Creates RegularGridInterpolator objects for each emission species using the training data.
    The interpolators are stored in the emissions.surrogates Data structure.

    The surrogate models interpolate emission indices based on:

    - Pressure
    - Temperature 
    - Air mass flow rate
    - Fuel-to-air ratio

    **Extra modules required**

    * scipy.interpolate

    **Major Assumptions**

    * Linear interpolation is sufficient between training points

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.train_CRN_EI_surrogates
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_CRN_emission_indices
    """
     
    surrogates                            = emissions.surrogates
    training                              = emissions.training  
    pressure_data                         = training.pressure         
    temperature_data                      = training.temperature      
    air_mass_flowrate_data                = training.air_mass_flowrate
    fuel_to_air_ratio_data                = training.fuel_to_air_ratio

    surrogates.EI_CO2                     = RegularGridInterpolator((pressure_data ,temperature_data, air_mass_flowrate_data, fuel_to_air_ratio_data),training.EI_CO2 ,method = 'linear',   bounds_error=False, fill_value=None) 
    surrogates.EI_CO                      = RegularGridInterpolator((pressure_data ,temperature_data, air_mass_flowrate_data, fuel_to_air_ratio_data),training.EI_CO  ,method = 'linear',   bounds_error=False, fill_value=None) 
    surrogates.EI_H2O                     = RegularGridInterpolator((pressure_data ,temperature_data, air_mass_flowrate_data, fuel_to_air_ratio_data),training.EI_H2O ,method = 'linear',   bounds_error=False, fill_value=None) 
    surrogates.EI_NO                      = RegularGridInterpolator((pressure_data ,temperature_data, air_mass_flowrate_data, fuel_to_air_ratio_data),training.EI_NO  ,method = 'linear',   bounds_error=False, fill_value=None) 
    surrogates.EI_NO2                     = RegularGridInterpolator((pressure_data ,temperature_data, air_mass_flowrate_data, fuel_to_air_ratio_data),training.EI_NO2 ,method = 'linear',   bounds_error=False, fill_value=None) 
   
    return


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/supersonic_wave_drag_volume_sears_haack.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# package imports
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Supersonic Wave Drag Due to Volume - Sears Haack
# ---------------------------------------------------------------------------------------------------------------------- 
def supersonic_wave_drag_volume_sears_haack(vehicle, mach, scaling_factor, sears_haack_type = 3):
    """Computes the volume drag
    
    Assumptions:
    Basic fit
    
    Source:
    Sieron, Thomas R., et al. Procedures and design data for the formulation of aircraft 
    configurations. WRIGHT LAB WRIGHT-PATTERSON AFB OH, 1993. Page B-3
    
    Args:
    vehicle.
      total_length                        [m]
      maximum_cross_sectional_area        [m^2]
      reference_area                      [m^2]
      
    Returns:
    vehicle_wave_drag                     [Unitless] 
    """

    L        = vehicle.total_length
    Amax     = vehicle.maximum_cross_sectional_area
    S        = vehicle.reference_area

    rmax     = np.sqrt(Amax/np.pi)
    d        = rmax*2

    if sears_haack_type == 3:
        # Compute drag from Sears-Haack type III body
        # Source formula uses front projected area as a reference
        CD = 3/2*np.pi*np.pi*(d/L)*(d/L)*Amax/S
    else:
        raise NotImplementedError # can add other Sears-Haack types here
        
    # Scale to account for non-ideal shaping
    CD_scaled = CD*scaling_factor

    return CD_scaled 



# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

#RCAIDE Imports
from RCAIDE import  * 
from RCAIDE.Framework.Core import Units,  Data
from RCAIDE.Library.Components import Wings

from RCAIDE.Library.Methods.Aerodynamics.Common.Lift.compute_slat_lift import compute_slat_lift
from RCAIDE.Library.Methods.Aerodynamics.Common.Lift.compute_flap_lift import compute_flap_lift

# ----------------------------------------------------------------------
#  compute_max_lift_coeff
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Fidelity_Zero-Lift
def compute_max_lift_coeff(state,settings,geometry):
    """Computes the maximum lift coefficient associated with an aircraft high lift system

    Assumptions:
    None

    Source:
    Unknown

    Inputs:
    analyses.max_lift_coefficient_factor       [Unitless]
    vehicle.reference_area                     [m^2]
    vehicle.wings.                             
      areas.reference                          [m^2]
      thickness_to_chord                       [Unitless]
      chords.mean_aerodynamic                  [m]
      sweeps.quarter_chord                     [radians]
      taper                                    [Unitless]
      flaps.chord                              [m]
     control_surfaces.flap.deflection          [radians]
     control_surfaces.slat.deflection          [radians]
      areas.affected                           [m^2]
      control_surfaces.flap.configuration_type [string]
    conditions.freestream.                     
      velocity                                 [m/s]
      density                                  [kg/m^3]
      dynamic_viscosity                        [N s/m^2]
                                               
    Outputs:                                   
    Cl_max_ls (maximum CL)                     [Unitless]
    Cd_ind    (induced drag)                   [Unitless]

    Properties Used:
    N/A
    """    


    # initializing Cl and CDi
    Cl_max_ls = 0
    Cd_ind    = 0
    vehicle = geometry
    conditions = state.conditions

    #unpack
    max_lift_coefficient_factor = settings.maximum_lift_coefficient_factor
    for wing in vehicle.wings:
    
        if not wing.high_lift: continue
        #geometrical data
        Sref       = vehicle.reference_area
        Swing      = wing.areas.reference
        tc         = wing.thickness_to_chord * 100
        chord_mac  = wing.chords.mean_aerodynamic
        sweep      = wing.sweeps.quarter_chord
        sweep_deg  = wing.sweeps.quarter_chord / Units.degree # convert into degrees
        taper      = wing.taper
        
        # conditions data
        V    = conditions.freestream.velocity
        roc  = conditions.freestream.density
        nu   = conditions.freestream.dynamic_viscosity

        #--cl max based on airfoil t_c
        Cl_max_ref = -0.0009*tc**3 + 0.0217*tc**2 - 0.0442*tc + 0.7005
        #-reynolds number effect
        Reyn     =  V * roc * chord_mac / nu
        Re_ref   = 9*10**6
        op_Clmax = Cl_max_ref * ( Reyn / Re_ref ) **0.1

        #wing cl_max to outer panel Cl_max
        w_Clmax = op_Clmax* ( 0.919729714285715 -0.044504761904771*taper \
                             -0.001835900000000*sweep_deg +  0.247071428571446*taper**2 +  \
                              0.003191500000000*taper*sweep_deg -0.000056632142857*sweep_deg**2  \
                             -0.279166666666676*taper**3 +  0.002300000000000*taper**2*sweep_deg + \
                              0.000049982142857*taper*sweep_deg**2  -0.000000280000000* sweep_deg**3)

        #---FAR stall speed effect---------------
        #should be optional based on aircraft being modelled
        Cl_max_FAA = 1.1 * w_Clmax

        #-----------wing mounted engine ----
        Cl_max_w_eng = Cl_max_FAA - 0.2

        # Compute CL increment due to Flap
        if 'slat' in wing.control_surfaces.keys():
         # Compute CL increment due to Slat
            slat_angle = wing.control_surfaces.slat.deflection
            dcl_slat = compute_slat_lift(slat_angle, sweep)
        else:
            dcl_slat = 0.

        # Compute CL increment due to Flap
        if 'flap' in wing.control_surfaces.keys():
            flap_type  = wing.control_surfaces.flap.configuration_type
            flap_chord = wing.control_surfaces.flap.chord_fraction # correct !!! 
            flap_angle = wing.control_surfaces.flap.deflection
            Swf        = wing.areas.affected  # portion of wing area with flaps
            dcl_flap   = compute_flap_lift(tc,flap_type,flap_chord,flap_angle,sweep,Sref,Swf)
        else:
            dcl_flap = 0.0

        #results
        Cl_max_ls += (Cl_max_w_eng + dcl_slat + dcl_flap) * Swing / Sref
        Cd_ind += ( 0.01 ) * Swing / Sref

    Cl_max_ls = Cl_max_ls * max_lift_coefficient_factor
    return Cl_max_ls, Cd_ind



# RCAIDE/Library/Methods/Propulsors/Common/compute_static_sea_level_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jul 2024, RCAIDE Team

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE  
from RCAIDE.Framework.Mission.Common      import  Conditions

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Design Turbofan
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_static_sea_level_performance(propulsor):
    """Compute static sea level performance of a propulsor     
    
    Assumtions:
       None 
    
    Source:
        None 
    
    Args:
        propulsor (dict): propulsor data structure [-]
    
    Returns:
        None 
    
    """  
    # Step 28: Static Sea Level Thrust  
    planet                                            = RCAIDE.Library.Attributes.Planets.Earth()
    atmosphere_sls                                    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data                                         = atmosphere_sls.compute_values(0.0,0.0)
                                                      
    p                                                 = atmo_data.pressure          
    T                                                 = atmo_data.temperature       
    rho                                               = atmo_data.density          
    a                                                 = atmo_data.speed_of_sound    
    mu                                                = atmo_data.dynamic_viscosity     
                                                      
    conditions                                        = RCAIDE.Framework.Mission.Common.Results() 
    conditions.freestream.altitude                    = np.atleast_1d(0)
    conditions.freestream.mach_number                 = np.atleast_1d(0.01)
    conditions.freestream.pressure                    = np.atleast_1d(p)
    conditions.freestream.temperature                 = np.atleast_1d(T)
    conditions.freestream.density                     = np.atleast_1d(rho)
    conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
    conditions.freestream.gravity                     = np.atleast_2d(planet.sea_level_gravity)
    conditions.freestream.isentropic_expansion_factor = np.atleast_1d(propulsor.working_fluid.compute_gamma(T,p))
    conditions.freestream.Cp                          = np.atleast_1d(propulsor.working_fluid.compute_cp(T,p))
    conditions.freestream.R                           = np.atleast_1d(propulsor.working_fluid.gas_specific_constant)
    conditions.freestream.speed_of_sound              = np.atleast_1d(a)
    conditions.freestream.velocity                    = np.atleast_1d(a*0.01)   

    # setup conditions   
    segment                                           = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions                          = conditions      
    propulsor.append_operating_conditions(segment) 
    for tag, item in  propulsor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,propulsor) 
    
    # set throttle
    segment.state.conditions.energy[propulsor.tag].throttle[:,0] = 1.0  
    sls_T,sls_M,sls_P,_,_                             = propulsor.compute_performance(segment.state)
        
    propulsor.sealevel_static_thrust                  = sls_T[0][0]
    return 


# RCAIDE/Library/Methods/Thermal_Management/Batteries/Air_Cooled/append_air_cooled_conditions.py
# 
# Created:  Aug 2024, S. Shekar


# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_turbofan_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_air_cooled_conditions(air_cooled,segment,coolant_line):
    """ Packs the initial air cooling heat acqusition conditions
    
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:  
             air_cooled          (data structure)              [None]
           
        Outputs:
            segment
              effectiveness                                    [boolean]
              total_heat_removed                               [watts]
   
        Properties Used:
        None
    """    
    
    ones_row                                                                                        = segment.state.ones_row                  
    segment.state.conditions.energy[coolant_line.tag][air_cooled.tag]                               = Conditions()
    segment.state.conditions.energy[coolant_line.tag][air_cooled.tag].effectiveness                 = 0. * ones_row(1)
    segment.state.conditions.energy[coolant_line.tag][air_cooled.tag].total_heat_removed            = 0. * ones_row(1)
    segment.state.conditions.energy[coolant_line.tag][air_cooled.tag].power                         = 0. * ones_row(1)
    
    return

def append_air_cooled_segment_conditions(air_cooled, segment,coolant_line, conditions):
    """Sets the initial air cooled heat acquistion conditions at the start of each segment as the last point from the previous segment 
    
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:  
         air_cooled          (data structure)              [None]
               
        Outputs:
        None
    
        Properties Used:
        None
    """    
    
    air_cooled_conditions = conditions[coolant_line.tag][air_cooled.tag]
    if segment.state.initials:  
        air_cooled_initials                            = segment.state.initials.conditions.energy[coolant_line.tag][air_cooled.tag]
        air_cooled_conditions.effectiveness[:,0]       = air_cooled_initials.effectiveness[-1,0]     
        air_cooled_conditions.total_heat_removed[:,0]  = air_cooled_initials.total_heat_removed[-1,0]   
    
    return


# RCAIDE/Library/Missions/Common/Initialize/energy.py
# 
# 
# Created:  Jul 2023, M. Clarke
# Modified: Sep 2024, S. Shekar


# ----------------------------------------------------------------------------------------------------------------------
#  energy
# ----------------------------------------------------------------------------------------------------------------------  
def energy(segment):
    """ Sets the initial battery energy at the start of the mission

        Assumptions:
        N/A

        Inputs:
            segment
            
        Outputs:
            None
           
        Properties Used:
        N/A

    """ 

    conditions = segment.state.conditions.energy
    
    # loop through battery modules in networks
    for network in segment.analyses.energy.vehicle.networks:  
        # if network has busses 
        if 'busses' in network:
            for bus in network.busses:
                bus.append_segment_conditions(conditions, segment)
                for battery_module in  bus.battery_modules:
                    battery_module.append_battery_segment_conditions(bus, conditions, segment)
                for coolant_line in  network.coolant_lines:
                    for tag, item in  coolant_line.items(): 
                        if tag == 'battery_modules':
                            for battery in item:
                                for btms in  battery:
                                    btms.append_segment_conditions(segment,coolant_line, conditions)
                        if tag == 'heat_exchangers':
                            for heat_exchanger in  item:
                                heat_exchanger.append_segment_conditions(segment,bus,coolant_line, conditions)
                        if tag == 'reservoirs':
                            for reservoir in  item:
                                reservoir.append_segment_conditions(segment, coolant_line, conditions) 
                    
         # if network has fuel lines                        
        elif 'fuel_lines' in network: 
            for fuel_line in  network.fuel_lines:
                for fuel_tank in fuel_line.fuel_tanks: 
                    fuel_tank_conditions   = conditions[fuel_line.tag][fuel_tank.tag] 
                    if segment.state.initials:  
                        fuel_tank_initials = segment.state.initials.conditions.energy[fuel_line.tag][fuel_tank.tag] 
                        fuel_tank_conditions.mass[:,0]   = fuel_tank_initials.mass[-1,0]
                    else: 
                        fuel_tank_conditions.mass[:,0]   = segment.analyses.energy.vehicle.networks[network.tag].fuel_lines[fuel_line.tag].fuel_tanks[fuel_tank.tag].fuel.mass_properties.mass


# RCAIDE/Library/Components/Energy/Sources/Battery_Modules/__init__.py
# 

"""
Battery module components for aircraft energy storage systems

This module provides implementations of various battery technologies including
lithium-ion, lithium-sulfur, lithium-air, and aluminum-air batteries. Each battery
type has specific characteristics for energy density, power density, and lifecycle
performance.
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Aluminum_Air           import Aluminum_Air
from .Generic_Battery_Module import Generic_Battery_Module
from .Lithium_Ion_LFP        import Lithium_Ion_LFP
from .Lithium_Ion_NMC        import Lithium_Ion_NMC  
from .Lithium_Sulfur         import Lithium_Sulfur
from .Lithium_Air            import Lithium_Air 



## @ingroup Methods-Performance
# estimate_take_off_field_length.py
#
# Created:  Jun 2014, T. Orra, C. Ilario, Celso, 
# Modified: Apr 2015, M. Vegh 
#           Jan 2016, E. Botero
#           Mar 2020, M. Clarke
#           May 2020, E. Botero
#           Jul 2020, E. Botero 


# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE
from RCAIDE.Framework.Core            import Data, Units     
from RCAIDE.Library.Methods.Aerodynamics.Common.Drag import * 
from RCAIDE.Library.Methods.Aerodynamics.Common.Lift import *

# package imports
import numpy as np

# ----------------------------------------------------------------------
#  Compute field length required for takeoff
# ----------------------------------------------------------------------

## @ingroup Methods-Performance
def estimate_take_off_field_length(vehicle,analyses,altitude = 0, delta_isa = 0, compute_2nd_seg_climb = False):
    """ Computes the takeoff field length for a given vehicle configuration in a given airport.
    Also optionally computes the second segment climb gradient.

    Assumptions:
    For second segment climb gradient:
    One engine inoperative
    Only validated for two engine aircraft

    Source:
    http://adg.stanford.edu/aa241/AircraftDesign.html

    Inputs:
    analyses.base.atmosphere               [RCAIDE data type]
    airport.
      altitude                             [m]
      delta_isa                            [K]
    vehicle.
      mass_properties.takeoff              [kg]
      reference_area                       [m^2]
      V2_VS_ratio (optional)               [Unitless]
      maximum_lift_coefficient (optional)  [Unitless]
      networks.*.number_of_engines       [Unitless]

    Outputs:
    takeoff_field_length                   [m]

    Properties Used:
    N/A
    """        

    # ==============================================
        # Unpack
    # ==============================================
    atmo            = analyses.atmosphere 
    weight          = vehicle.mass_properties.takeoff
    reference_area  = vehicle.reference_area
    try:
        V2_VS_ratio = vehicle.V2_VS_ratio
    except:
        V2_VS_ratio = 1.20
        
        
    

    # ==============================================
    # Computing atmospheric conditions
    # ==============================================
    atmo_values       = atmo.compute_values(altitude,delta_isa)
    conditions        = RCAIDE.Framework.Mission.Common.Results()
    
    p   = atmo_values.pressure
    T   = atmo_values.temperature
    rho = atmo_values.density
    a   = atmo_values.speed_of_sound
    mu  = atmo_values.dynamic_viscosity
    sea_level_gravity = atmo.planet.sea_level_gravity
    
    # ==============================================
    # Determining vehicle maximum lift coefficient
    # ==============================================
    # Condition to CLmax calculation: 90KTAS @ airport
    state = Data()
    state.conditions = RCAIDE.Framework.Mission.Common.Results()
    state.conditions.freestream = Data()
    state.conditions.freestream.density           = rho
    state.conditions.freestream.velocity          = 90. * Units.knots
    state.conditions.freestream.dynamic_viscosity = mu
    
    settings = analyses.aerodynamics.settings

    maximum_lift_coefficient, induced_drag_high_lift = compute_max_lift_coeff(state,settings,vehicle)

    # ==============================================
    # Computing speeds (Vs, V2, 0.7*V2)
    # ==============================================
    stall_speed       = (2 * weight * sea_level_gravity / (rho * reference_area * maximum_lift_coefficient)) ** 0.5
    V2_speed          = V2_VS_ratio * stall_speed
    speed_for_thrust  = 0.70 * V2_speed

    # ==============================================
    # Determining vehicle number of engines
    # ==============================================
    engine_number = 0.
    for network in vehicle.networks : # may have than one network
        engine_number += len(network.propulsors)
    if engine_number == 0:
        raise ValueError("No engine found in the vehicle")

    # ==============================================
    # Getting engine thrust
    # ==============================================
    

    # Step 28: Static Sea Level Thrust  
    planet                                            = RCAIDE.Library.Attributes.Planets.Earth()
    atmosphere_sls                                    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data                                         = atmosphere_sls.compute_values(0.0,0.0)
                                                      
    p                                                 = atmo_data.pressure          
    T                                                 = atmo_data.temperature       
    rho                                               = atmo_data.density          
    a                                                 = atmo_data.speed_of_sound    
    mu                                                = atmo_data.dynamic_viscosity 
    
    conditions                                        = RCAIDE.Framework.Mission.Common.Results() 
    conditions.freestream.altitude                    = np.atleast_1d(0)
    conditions.freestream.mach_number                 = np.atleast_1d(0.01)
    conditions.freestream.pressure                    = np.atleast_1d(p)
    conditions.freestream.temperature                 = np.atleast_1d(T)
    conditions.freestream.density                     = np.atleast_1d(rho)
    conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
    conditions.freestream.gravity                     = np.atleast_2d(planet.sea_level_gravity) 
    conditions.freestream.speed_of_sound              = np.atleast_1d(a)
    conditions.freestream.velocity                    = np.atleast_1d(a*0.01)   

    # setup conditions   
    segment                                           = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions                          = conditions
    
    thrust =  np.array([[0.0, 0.0, 0.0]])
     
    analysis = RCAIDE.Framework.Analyses.Vehicle() 
    energy   = RCAIDE.Framework.Analyses.Energy.Energy()
    energy.vehicle  = vehicle 
    analysis.append(energy)            
    segment.analyses = analysis      
    
    for network in vehicle.networks:
        network.add_unknowns_and_residuals_to_segment(segment)
        
        for propulsor in  network.propulsors: 
            segment.state.conditions.energy[propulsor.tag].throttle = np.array([[1]])
        
        network.evaluate(segment.state,center_of_gravity = vehicle.mass_properties.center_of_gravity)
        
        thrust += conditions.energy.thrust_force_vector
         
    # ==============================================
    # Calculate takeoff distance
    # ==============================================

    # Defining takeoff distance equations coefficients 
    takeoff_constants = np.zeros(3)
    if engine_number == 2:
        takeoff_constants[0] =   857.4
        takeoff_constants[1] =   2.476
        takeoff_constants[2] =   0.00014
    elif engine_number == 3:
        takeoff_constants[0] = 667.9
        takeoff_constants[1] =   2.343
        takeoff_constants[2] =   0.000093
    elif engine_number == 4:
        takeoff_constants[0] = 486.7
        takeoff_constants[1] =   2.282
        takeoff_constants[2] =   0.0000705
    elif engine_number >  4:
        takeoff_constants[0] = 486.7
        takeoff_constants[1] =   2.282
        takeoff_constants[2] =   0.0000705
        print('The vehicle has more than 4 engines. Using 4 engine correlation. Result may not be correct.')
    else:
        takeoff_constants[0] = 857.4
        takeoff_constants[1] =   2.476
        takeoff_constants[2] =   0.00014
        print('Incorrect number of engines: {0:.1f}. Using twin engine correlation.'.format(engine_number))

    # Define takeoff index   (V2^2 / (T/W)
    takeoff_index = V2_speed**2. / (thrust[0][0] / weight)
    
    # Calculating takeoff field length
    takeoff_field_length = 0.
    for idx,constant in enumerate(takeoff_constants):
        takeoff_field_length += constant * takeoff_index**idx
    takeoff_field_length = takeoff_field_length * Units.ft
    
    # calculating second segment climb gradient, if required by user input
    if compute_2nd_seg_climb:
        
        # Getting engine thrust at V2 (update only speed related conditions)
        state.conditions.freestream.dynamic_pressure  = np.array(np.atleast_1d(0.5 * rho * V2_speed**2))
        state.conditions.freestream.velocity          = np.array(np.atleast_1d(V2_speed))
        state.conditions.freestream.mach_number       = np.array(np.atleast_1d(V2_speed/ a))
        state.conditions.freestream.dynamic_viscosity = np.array(np.atleast_1d(mu))
        state.conditions.freestream.density           =  np.array(np.atleast_1d(rho))
        
        # engine condition
        num_propulsors =  0
        for network in vehicle.networks:
            num_propulsors += len(network.propulsors) 
            for propulsor in  network.propulsors: 
                engine_out_location = propulsor.origin[0][1] 
        thrust  = thrust * (num_propulsors -1 )/num_propulsors
        single_engine_thrust =  np.linalg.norm(thrust /num_propulsors)

        # Compute windmilling drag
        windmilling_drag_coefficient = windmilling_drag(vehicle,state)

        # Compute asymmetry drag   
        asymmetry_drag_coefficient = asymmetry_drag(state, vehicle,engine_out_location, single_engine_thrust, windmilling_drag_coefficient)
           
        # Compute l over d ratio for takeoff condition, NO engine failure
        l_over_d = estimate_2ndseg_lift_drag_ratio(state,settings,vehicle) 
        
        # Compute L over D ratio for takeoff condition, WITH engine failure
        clv2            = maximum_lift_coefficient / (V2_VS_ratio) **2
        cdv2_all_engine = clv2 / l_over_d
        cdv2            = cdv2_all_engine + asymmetry_drag_coefficient + windmilling_drag_coefficient
        l_over_d_v2     = clv2 / cdv2
    
        # Compute 2nd segment climb gradient
        second_seg_climb_gradient = thrust / (weight*sea_level_gravity) - 1. / l_over_d_v2

        return takeoff_field_length[0][0], second_seg_climb_gradient[0][0]

    else:
        # return only takeoff_field_length
        return takeoff_field_length[0][0],0


# RCAIDE/Methods/Noise/Frequency_Domain_Buildup/Rotor/TIP_broadband_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Python Package imports  
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
# Compute TIP Broadband Noise 
# ---------------------------------------------------------------------------------------------------------------------- 
def TIP_broadband_noise(alpha_tip,M,c,c_0,f,Dbar_h,r_e):
    '''This computes the tip noise compoment of broadband noise using the method outlined by the 
    Brooks, Pope and Marcolini (BPM) Model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol  
        
    Source:  
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs:  
       alpha_TIP - angle of attack of tip section         [deg]
       M         - Mach number                            [-]
       c         - airfoil section chord                  [m]
       c_0       - speed of sound                         [m/s]
       f         - frequency spectrum                     [Hz]
       Dbar_h    - high frequency directivity term        [-]
       r_e       - distance from noise source to observer [m] 
    
    Outputs 
       SPL_TIP   - Sound pressure level of tip            [dB]
       
    Properties Used:
        N/A   
    '''       
    l_div_c                = 0.023 + 0.0169*alpha_tip              # eqn 67 BPM Paper 
    l_div_c[2<alpha_tip]   = 0.0378 + 0.0095*alpha_tip[2<alpha_tip]
    l                      = l_div_c*c                    # eqn 63 BPM Paper 
    M_max_div_M            = (1 + 0.036*alpha_tip)        # eqn 64 BPM Paper
    M_max                  = M_max_div_M * M              # eqn 64 BPM Paper
    U_max                  = c_0*M_max 
    St_prime_prime         = f*l/U_max                    # eqn 62 BPM Paper 
    SPL_TIP                = 10*np.log10(((M**2)*(M_max**3)*(l**2)*Dbar_h)/(r_e**2)) - 30.5*(np.log10(St_prime_prime + 0.3))**2 + 126 # eqn 61  
    return SPL_TIP



# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from .aero_coeff             import aero_coeff    
from .airfoil_analysis       import airfoil_analysis 
from .heads_method           import heads_method       
from .hess_smith             import hess_smith               
from .infl_coeff             import infl_coeff       
from .panel_geometry         import panel_geometry   
from .thwaites_method        import thwaites_method    
from .velocity_distribution  import velocity_distribution
from .cf_filter              import cf_filter
from .chordwise_distribution import chordwise_distribution


# RCAIDE/Library/Missions/Common/Pre_Process/set_residuals_and_unknowns.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units

# ----------------------------------------------------------------------------------------------------------------------
#  set_residuals_and_unknowns
# ----------------------------------------------------------------------------------------------------------------------  
def set_residuals_and_unknowns(mission):
    """ Sets the flight dynamics residuals and fligth controls of the aircraft   

        Assumptions:
        N/A

        Inputs: 
            mission     - data structure of mission  [-]
 
        Outputs:  

        Properties Used:
        N/A

    """     
    for segment in mission.segments:  
        ones_row    = segment.state.ones_row 
        ctrls       = segment.assigned_control_variables
        dynamics    = segment.flight_dynamics
        
        # assign force and moment residuals i.e. degrees of freedom 
        num_DOF  = 0      
        if dynamics.force_x == True: 
            segment.state.residuals.force_x = ones_row(1) *0 
            num_DOF += 1 
        if dynamics.force_y == True:
            segment.state.residuals.force_y = ones_row(1) *0
            num_DOF += 1  
        if dynamics.force_z == True:
            segment.state.residuals.force_z = ones_row(1) *0
            num_DOF += 1  
        if dynamics.moment_x == True:
            segment.state.residuals.moment_x = ones_row(1) *0
            num_DOF += 1  
        if dynamics.moment_y == True:
            segment.state.residuals.moment_y = ones_row(1) *0 
            num_DOF += 1 
        if dynamics.moment_z == True:
            segment.state.residuals.moment_z = ones_row(1) *0
            num_DOF += 1  
        
        # assign control variables   
        ones_row     = segment.state.ones_row 
        num_ctrls    = 0
        
        # Body Angle  
        if ctrls.body_angle.active:
            if ctrls.body_angle.initial_guess_values !=  None:
                segment.state.unknowns.body_angle = ones_row(1) * ctrls.body_angle.initial_guess_values[0][0]
            else:
                segment.state.unknowns.body_angle = ones_row(1) * 3.0 * Units.degrees 
            num_ctrls += 1  
    
        # Bank Angle  
        if ctrls.bank_angle.active:
            if ctrls.bank_angle.initial_guess_values !=  None:
                segment.state.unknowns.bank_angle = ones_row(1) * ctrls.bank_angle.initial_guess_values[0][0]
            else:
                segment.state.unknowns.bank_angle = ones_row(1) * 0.0 * Units.degrees 
            num_ctrls += 1     
                
        # Wing Angle  
        if ctrls.wind_angle.active:
            if ctrls.wind_angle.initial_guess_values !=  None:
                segment.state.unknowns.wind_angle = ones_row(1) * ctrls.wind_angle.initial_guess_values[0][0]
            else:
                segment.state.unknowns.wind_angle = ones_row(1) * 1.0 * Units.degrees 
            num_ctrls += 1            
            
        # Throttle 
        if ctrls.throttle.active: 
            for i in range(len(ctrls.throttle.assigned_propulsors)): 
                if ctrls.throttle.initial_guess_values !=  None:
                    segment.state.unknowns["throttle_" + str(i)] = ones_row(1) * ctrls.throttle.initial_guess_values[i][0] 
                else:
                    segment.state.unknowns["throttle_" + str(i)] = ones_row(1) *  0.5
                num_ctrls += 1    
        
        # Velocity 
        if ctrls.velocity.active:  
            if  ctrls.velocity.initial_guess_values !=  None:
                segment.state.unknowns.velocity = ones_row(1) * ctrls.velocity.initial_guess_values[0][0] 
            else:
                segment.state.unknowns.velocity = ones_row(1) *  100
            num_ctrls += 1    
                
        # Acceleration 
        if ctrls.acceleration.active:  
            if ctrls.acceleration.initial_guess_values !=  None:
                segment.state.unknowns.acceleration = ones_row(1) * ctrls.acceleration.initial_guess_values[0][0] 
            else:
                segment.state.unknowns.acceleration = ones_row(1) *  1.
            num_ctrls += 1   

        # Time
        if ctrls.elapsed_time.active:  
            if ctrls.elapsed_time.initial_guess_values != None: 
                segment.state.unknowns.elapsed_time = ctrls.elapsed_time.initial_guess_values[0][0] 
            else:
                segment.state.unknowns.elapsed_time = 10
            num_ctrls += 1                         
                                
        # Elevator 
        if ctrls.elevator_deflection.active:     
            for i in range(len(ctrls.elevator_deflection.assigned_surfaces)): 
                if ctrls.elevator_deflection.initial_guess_values!= None:  
                    segment.state.unknowns["elevator_" + str(i)] = ones_row(1) * ctrls.elevator_deflection.initial_guess_values[i][0]
                else:
                    segment.state.unknowns["elevator_" + str(i)] = ones_row(1) * 0.0 * Units.degrees  
                num_ctrls += 1   
                
        # Elevator 
        if ctrls.rudder_deflection.active:  
            for i in range(len(ctrls.rudder_deflection.assigned_surfaces)):   
                if ctrls.rudder_deflection.initial_guess_values !=  None: 
                    segment.state.unknowns["rudder_" + str(i)] = ones_row(1) * ctrls.rudder_deflection.initial_guess_values[i][0]
                else:
                    segment.state.unknowns["rudder_" + str(i)] = ones_row(1) * 0.0 * Units.degrees  
                num_ctrls += 1    
                    
        # Flap  
        if ctrls.flap_deflection.active:  
            for i in range(len(ctrls.flap_deflection.assigned_surfaces)):
                if ctrls.flap_deflection.initial_guess_values !=  None:
                    segment.state.unknowns["flap_" + str(i)] = ones_row(1) * ctrls.flap_deflection.initial_guess_values[i][0]
                else:
                    segment.state.unknowns["flap_" + str(i)] = ones_row(1) * 0.0 * Units.degrees 
                num_ctrls += 1    
        # Slat  
        if ctrls.slat_deflection.active:  
            for i in range(len(ctrls.slat_deflection.assigned_surfaces)):  
                if ctrls.slat_deflection.initial_guess_values != None:      
                    segment.state.unknowns["slat_" + str(i)] = ones_row(1) * ctrls.slat_deflection.initial_guess_values[i][0]
                else:
                    segment.state.unknowns["slat_" + str(i)] = ones_row(1) * 0.0 * Units.degrees 
                num_ctrls += 1   
                
        # Aileron  
        if ctrls.aileron_deflection.active:  
            for i in range(len(ctrls.aileron_deflection.assigned_surfaces)):   
                if ctrls.aileron_deflection.initial_guess_values !=  None:
                    segment.state.unknowns["aileron_" + str(i)] = ones_row(1) * ctrls.aileron_deflection.initial_guess_values[i][0]
                else: 
                    segment.state.unknowns["aileron_" + str(i)] = ones_row(1) * 0.0 * Units.degrees 
                num_ctrls += 1       
            
        #  Thrust Vector Angle
        if ctrls.thrust_vector_angle.active:  
            for i in range(len(ctrls.thrust_vector_angle.assigned_propulsors)):  
                if ctrls.thrust_vector_angle.initial_guess_values != None:  
                    segment.state.unknowns["thrust_vector_" + str(i)] = ones_row(1) * ctrls.thrust_vector_angle.initial_guess_values[i][0]
                else:
                    segment.state.unknowns["thrust_vector_" + str(i)] = ones_row(1) * 0.0 * Units.degrees 
                num_ctrls += 1         
        
        # TO DO: add pitch command
        
        # TO DO: if the degrees of freedom are greater than the number of control inputs, post problem at optimization   
    return 
                                                                                                                                                                


# RCAIDE/Methods/Energy/Propulsors/Converters/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Combustor
from . import Compression_Nozzle
from . import Compressor
from . import DC_Motor
from . import Ducted_Fan
from . import Engine
from . import Expansion_Nozzle
from . import Supersonic_Nozzle
from . import Fan 
from . import Ram
from . import Rotor 
from . import Offtake_Shaft
from . import Turbine


# RCAIDE/Library/Methods/Propulsors/Converters/Combustor/__init__.py
# (c) Copyright 2023 Aerospace Research Community LLC

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
        
from .compute_combustor_performance         import compute_combustor_performance 
from .compute_rayleigh                      import compute_rayleigh
from .compute_supersonic_combustion         import compute_supersonic_combustion
from .append_combustor_conditions           import append_combustor_conditions 


# RCAIDE/Library/Plots/Thermal_Management/plot_thermal_management_performance.py
# 
# 
# Created:  Sep 2024, S. Shekar
# ----------------------------------------------------------------------------------------------------------------------
#   plot_thermal_management_performance
# ----------------------------------------------------------------------------------------------------------------------   
def plot_thermal_management_performance(results,
                        save_figure   = False,
                        show_legend   = True,
                        file_type     =".png",
                        width         = 12,
                        height        = 7):
    """Checks and plots all componenents of a thermal management system.
    
     Assumptions:
     None
    
     Source:
     None
    
     Inputs:
     results
     Outputs:
     Plots
    
     Properties Used:
     N/A	
     """     
    
    for network in  results.segments[0].analyses.energy.vehicle.networks:
        for coolant_line in  network.coolant_lines:
            for tag, item in  coolant_line.items():
                if coolant_line.identical_battery_modules:
                    if tag == 'battery_modules':
                            for i, battery in enumerate(item):
                                for btms in  (battery):
                                    if i ==  0:
                                        btms.plot_operating_conditions(results,coolant_line,save_figure,show_legend,btms.tag,file_type,width, height)
                else:
                    for _, battery in enumerate(item):
                            for btms in  (battery):
                                btms.plot_operating_conditions(results,coolant_line,save_figure,show_legend,btms.tag,file_type,width, height)
                if tag == 'heat_exchangers':
                    for heat_exchanger in  item:
                        heat_exchanger.plot_operating_conditions(results,coolant_line,save_figure,show_legend,heat_exchanger.tag,file_type,width, height)
                if tag == 'reservoirs':
                    for reservoir in  item:
                        reservoir.plot_operating_conditions(results,coolant_line,save_figure,show_legend,reservoir.tag,file_type,width, height)             
    return


# RCAIDE/Library/Missions/Segments/Transition/Constant_Acceleration_Constant_Angle_Linear_Climb.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  

import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------   
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant acceleration and constant altitude
    Velocities specified in the aircraft velocity frame (x-forward, z-down)

    Source:
    N/A

    Inputs:
    segment.altitude                            [meters]
    segment.air_speed_start                     [meters/second]
    segment.air_speed_end                       [meters/second]
    segment.acceleration                        [meters/second^2]
    conditions.frames.inertial.time             [seconds]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """      
    
    # unpack
    alt0        = segment.altitude_start 
    altf        = segment.altitude_end 
    climb_angle = segment.climb_angle
    v0          = segment.air_speed_start 
    beta        = segment.sideslip_angle    
    ax          = segment.acceleration   
    T0          = segment.pitch_initial
    Tf          = segment.pitch_final     
    t_nondim    = segment.state.numerics.dimensionless.control_points
    conditions  = segment.state.conditions 
    
    # check for climb angle     
    if climb_angle is None:
        raise AttributeError('set climb')
    
    if ax is None: 
        raise AttributeError('set acceleration') 
    
    # check for initial and final altitude 
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2] 
    
    if altf is None:
        raise AttributeError('final altitude not set')
        
    # check for initial pitch
    if T0 is None:
        T0  =  segment.state.initials.conditions.frames.body.inertial_rotations[-1,1] 
        
    # check for initial velocity vector
    if v0 is None:
        if not segment.state.initials: raise AttributeError('initial airspeed not set')
        v0  =  segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:] 
        segment.velocity_vector = v0
    
    elif len(np.shape(v0)) == 0:
        v0 = np.array([v0, 0, 0]) 
         
    # discretize on altitude
    v0_mag          = np.linalg.norm(v0)
    alt             = t_nondim * (altf-alt0) + alt0   
    ground_distance = (altf-alt0)/np.tan(climb_angle)
    true_distance   = np.sqrt((altf-alt0)**2 + ground_distance**2)
    t_initial       = conditions.frames.inertial.time[0,0]    
    elapsed_time    = (-v0_mag + np.sqrt(v0_mag**2 + 2*ax*true_distance))/(ax)   
    vf_mag          = v0_mag + ax*(elapsed_time)   
    
    # dimensionalize time        
    t_final   = t_initial + elapsed_time
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      = t_nondim * (t_final-t_initial)
    
    # Figure out vx
    V  = (vf_mag-v0_mag) 
    vx = t_nondim *  V  * np.cos(beta) * np.cos(climb_angle) + v0[0] * np.cos(climb_angle) 
    vy = t_nondim *  V  * np.sin(beta) * np.cos(climb_angle) + v0[1] * np.cos(climb_angle) 
    vz = t_nondim *  V  * np.sin(climb_angle) + v0[2] * np.sin(climb_angle)  
    
    # set the body angle
    body_angle = time*(Tf-T0)/(t_final-t_initial) + T0
    segment.state.conditions.frames.body.inertial_rotations[:,1] = body_angle[:,0]     
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt[:,0]
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = vx[:,0] 
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = vy[:,0] 
    segment.state.conditions.frames.inertial.velocity_vector[:,2] = -vz[:,0] 
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_propulsion_system_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
import  RCAIDE 
from RCAIDE.Framework.Core    import Units, Data
from RCAIDE.Library.Methods.Weights.Correlation_Buildups.FLOPS.compute_propulsion_system_weight import compute_engine_weight

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Propulsion System Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_propulsion_system_weight(vehicle,network):
    """ Calculate the weight of propulsion system using Raymer method, including:
        - fuel system weight
        - thurst reversers weight
        - electrical system weight
        - starter engine weight
        - nacelle weight
        - cargo containers
        The dry engine weight comes from the FLOPS relations since it is not listed in Raymer

        Assumptions:

        Source:
            Aircraft Design: A Conceptual Approach

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
            network    - data dictionary for the specific network that is being estimated [dimensionless]

        Outputs:
            output - data dictionary with weights                               [kilograms]
                    - output.W_prop: total propulsive system weight
                    - output.W_thrust_reverser: thurst reverser weight
                    - output.starter: starter engine weight
                    - output.W_engine_controls: engine controls weight
                    - output.fuel_system: fuel system weight
                    - output.nacelle: nacelle weight
                    - output.W_engine: dry engine weight

        Properties Used:
            N/A
    """

    NENG    =  0 
    number_of_tanks =  0
    for network in  vehicle.networks:
        for fuel_line in network.fuel_lines:
            for fuel_tank in fuel_line.fuel_tanks:
                number_of_tanks +=  1
            for propulsor in network.propulsors:
                if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbojet):
                    ref_propulsor = propulsor  
                    NENG  += 1 
                if 'nacelle' in propulsor:
                    ref_nacelle =  propulsor.nacelle 
                    
    WFSYS           = compute_fuel_system_weight(vehicle, NENG)
    WENG            = compute_engine_weight(vehicle,ref_propulsor)
    WNAC            = compute_nacelle_weight(vehicle,ref_nacelle, NENG, WENG)
    WEC, WSTART     = compute_misc_engine_weight(vehicle,NENG, WENG)
    WTHR            = 0
    WPRO            = NENG * WENG + WFSYS + WEC + WSTART + WTHR + WNAC

    output                      = Data()
    output.W_prop               = WPRO
    output.W_thrust_reverser    = WTHR
    output.W_starter            = WSTART
    output.W_engine_controls    = WEC
    output.W_fuel_system        = WFSYS
    output.W_nacelle            = WNAC
    output.W_engine             = WENG * NENG
    output.number_of_engines    = NENG
    output.number_of_fuel_tanks = number_of_tanks  
    return output

def compute_nacelle_weight(vehicle,ref_nacelle, NENG, WENG):
    """ Calculates the nacelle weight based on the Raymer method
        Assumptions:
            1) All nacelles are identical
            2) The number of nacelles is the same as the number of engines 
        Source:
            Aircraft Design: A Conceptual Approach (2nd edition)

        Inputs:
            vehicle - data dictionary with vehicle properties                           [dimensionless]
                -.ultimate_load: ultimate load factor of aircraft
            nacelle  - data dictionary for the specific nacelle that is being estimated [dimensionless]
                -lenght: total length of engine                                         [m]
                -diameter: diameter of nacelle                                          [m]
            WENG    - dry engine weight                                                 [kg]


        Outputs:
            WNAC: nacelle weight                                                        [kg]

        Properties Used:
            N/A
    """ 
    Kng             = 1 # assuming the engine is not pylon mounted
    Nlt             = ref_nacelle.length / Units.ft
    Nw              = ref_nacelle.diameter / Units.ft
    Wec             = 2.331 * WENG ** 0.901 * 1.18
    Sn              = 2 * np.pi * Nw/2 * Nlt + np.pi * Nw**2/4 * 2
    WNAC            = 0.6724 * Kng * Nlt ** 0.1 * Nw ** 0.294 * vehicle.flight_envelope.ultimate_load ** 0.119 \
                      * Wec ** 0.611 * NENG * 0.984 * Sn ** 0.224
    return WNAC * Units.lbs

def compute_misc_engine_weight(vehicle, NENG, WENG):
    """ Calculates the miscellaneous engine weight based on the Raymer method, electrical control system weight
        and starter engine weight
        Assumptions:

        Source:
            Aircraft Design: A Conceptual Approach

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.fuselages['fuselage'].lengths.total: length of fuselage   [m]
            network    - data dictionary for the specific network that is being estimated [dimensionless]
                -.number_of_engines: number of engines

        Outputs:
            WEC: electrical engine control system weight                    [kg]
            WSTART: starter engine weight                                   [kg]

        Properties Used:
            N/A
    """

    L =  0 
    for fuselage in vehicle.fuselages:
        if L < fuselage.lengths.total: 
            total_length = fuselage.lengths.total             
    Lec     = NENG * total_length / Units.ft
    WEC     = 5 * NENG + 0.8 * Lec
    WSTART  = 49.19*(NENG*WENG/1000)**0.541
    return WEC * Units.lbs, WSTART * Units.lbs
 
def compute_fuel_system_weight(vehicle, NENG):
    """ Calculates the weight of the fuel system based on the Raymer method
        Assumptions:

        Source:
            Aircraft Design: A Conceptual Approach

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.design_mach_number: design mach number
                -.mass_properties.max_zero_fuel: maximum zero fuel weight   [kg]

        Outputs:
            WFSYS: Fuel system weight                                       [kg]

        Properties Used:
            N/A
    """
    VMAX    = vehicle.flight_envelope.design_mach_number
    FMXTOT  = vehicle.mass_properties.max_zero_fuel / Units.lbs
    WFSYS = 1.07 * FMXTOT ** 0.58 * NENG ** 0.43 * VMAX ** 0.34
    return WFSYS * Units.lbs


## @ingroup Core
# Data.py
#
# Created:  Jun 2016, E. Botero
# Modified: Jan 2020, M. Clarke
#           May 2020, E. Botero
#           Jul 2021, E. Botero
#           Oct 2021, E. Botero



# ----------------------------------------------------------------------
#   Imports
# ----------------------------------------------------------------------

import numpy as np
from .Arrays import atleast_2d_col, array_type, matrix_type 

from copy import copy

# for enforcing attribute style access names
import string
from warnings import warn
chars = string.punctuation + string.whitespace
t_table = str.maketrans( chars          + string.ascii_uppercase , 
                            '_'*len(chars) + string.ascii_lowercase )

dictgetitem = dict.__getitem__
objgetattrib = object.__getattribute__

# ----------------------------------------------------------------------
#   Data
# ----------------------------------------------------------------------        

## @ingroup Core
class Data(dict):
    """ An extension of the Python dict which allows for both tag and '.' usage.
        This is an unordered dictionary. So indexing it will not produce deterministic results.
        This has less overhead than ordering. If ordering is needed use DataOrdered().
       
        Assumptions:
        N/A
        
        Source:
        N/A
    """
    
    def __getattribute__(self, k):
        """ Retrieves an attribute set by a key k
    
            Assumptions:
            Does a try if it is a dict, but if that fails treats it as an object
    
            Source:
            N/A
    
            Inputs:
            k
    
            Outputs:
            whatever is found by k
    
            Properties Used:
            N/A
            """         
        try:
            return dictgetitem(self,k)
        except:
            return objgetattrib(self,k)

    def __setattr__(self, k, v):
        """ An override of the standard __setattr_ in Python.
            
            Assumptions:
            This one tries to treat k as an object, if that fails it treats it as a key.
    
            Source:
            N/A
    
            Inputs:
            k        [key]
            v        [value]
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """
        try:
            objgetattrib(self, k)
        except:
            self[k] = v
        else:          
            object.__setattr__(self, k, v) 
            
    def __delattr__(self, k):
        """ An override of the standard __delattr_ in Python. This deletes whatever is called by k
            
            Assumptions:
            This one tries to treat k as an object, if that fails it treats it as a key.
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """        
        try:
            objgetattrib(self, k)
        except:
            del self[k]
        else:
            object.__delattr__(self, k)
    
    def __defaults__(self):
        """ A stub for all classes that come later
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """              
        pass      
    
    def __new__(cls,*args,**kwarg):
        """ Creates a new Data() class
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        
        
        # initialize data, no inputs
        self = super(Data,cls).__new__(cls)
        super(Data,self).__init__() 
        
        # get base class list
        klasses = self.get_bases()
                
        # fill in defaults trunk to leaf
        for klass in klasses[::-1]:
            try:
                klass.__defaults__(self)
            except:
                pass
            
        return self
    
    def typestring(self):
        """ This function makes the .key.key structure in string form of Data()
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        
        # build typestring
        typestring = str(type(self)).split("'")[1]
        typestring = typestring.split('.')
        if typestring[-1] == typestring[-2]:
            del typestring[-1]
        typestring = '.'.join(typestring) 
        return typestring    
    
    def dataname(self):
        """ This function is used for printing the class
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        return "<data object '" + self.typestring() + "'>"     
    
    
    def __str__(self,indent=''):
        """ This function is used for printing the class. This starts the first line of printing.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """         
        
        new_indent = '  '
        args = ''
        
        # trunk data name
        if not indent:
            args += self.dataname()  + '\n'
        else:
            args += ''
            
        args += self.__str2(indent)
        
        return args    
    
    
    def __str2(self,indent=''):
        """ This regresses through and does the rest of printing that __str__ missed
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """     
        
        new_indent = '  '
        args = ''
        
        # trunk data name
        if indent: args += '\n'
        
        # print values   
        for key,value in self.items():
            
            # skip 'hidden' items
            if isinstance(key,str) and key.startswith('_'):
                continue
            
            # recurse into other dict types
            if isinstance(value,dict):
                if not value:
                    val = '\n'
                else:
                    try:
                        val = value.__str2(indent+new_indent)
                    except RuntimeError: # recursion limit
                        val = ''
                    except:
                        val = value.__str__(indent+new_indent)
                                                
            # everything else
            else:
                val = str(value) + '\n'
                
            # this key-value, indented
            args+= indent + str(key) + ' : ' + val
            
        return args        
    
    def __init__(self,*args,**kwarg):
        """ :meta private:"""
        # """ 
        #     Initializes a new Data() class
    
        #     Assumptions:
        #     N/A
    
        #     Source:
        #     N/A
    
        #     Inputs:
        #     N/A
    
        #     Outputs:
        #     N/A
    
        #     Properties Used:
        #     N/A    
        # """           

        # handle input data (ala class factory)
        input_data = Data.__base__(*args,**kwarg)
        
        # update this data with inputs
        self.update(input_data)    

    def __iter__(self):
        """ :meta private:"""
        # """ Returns all the iterable values. Can be used in a for loop.
    
        #     Assumptions:
        #     N/A
    
        #     Source:
        #     N/A
    
        #     Inputs:
        #     N/A
    
        #     Outputs:
        #     N/A
    
        #     Properties Used:
        #     N/A    
        # """           
        return iter(self.values())
    
    def itervalues(self):
        """ :meta private:"""
        # """ Finds all the values that can be iterated over.
    
        #     Assumptions:
        #     N/A
    
        #     Source:
        #     N/A
    
        #     Inputs:
        #     N/A
    
        #     Outputs:
        #     N/A
    
        #     Properties Used:
        #     N/A    
        # """             
        for k in super(Data,self).__iter__():
            yield self[k]   
    
    def values(self):
        """ :meta private:"""
        # """ Returns all values inside the Data() class.
    
        #     Assumptions:
        #     N/A
    
        #     Source:
        #     N/A
    
        #     Inputs:
        #     N/A
    
        #     Outputs:
        #     values
    
        #     Properties Used:
        #     N/A    
        # """          
        return self.__values()          
            
    def __values(self):
        """ :meta private:"""
        #  Iterates over all keys to find all the data values.
    
        #     Assumptions:
        #     N/A
    
        #     Source:
        #     N/A
    
        #     Inputs:
        #     N/A
    
        #     Outputs:
        #     values
    
        #     Properties Used:
        #     N/A    
        #            
        return [self[key] for key in super(Data,self).__iter__()]    
    
    def update(self,other):
        """ :meta private:"""
        # """ Updates the internal values of a dictionary with given data
    
        #     Assumptions:
        #     N/A
    
        #     Source:
        #     N/A
    
        #     Inputs:
        #     other
    
        #     Outputs:
        #     N/A
    
        #     Properties Used:
        #     N/A    
        # """           
        if not isinstance(other,dict):
            raise TypeError('input is not a dictionary type')
        for k,v in other.items():
            # recurse only if self's value is a Dict()
            if k.startswith('_'):
                continue
        
            try:
                self[k].update(v)
            except:
                self[k] = v
        return 
    
    def get_bases(self):
        """ Finds the higher classes that may be built off of data
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            klasses
    
            Properties Used:
            N/A    
        """          
        # Get the Method Resolution Order, i.e. the ancestor tree
        klasses = list(self.__class__.__mro__)
        
        # Make sure that this is a Data object, otherwise throw an error.
        if Data not in klasses:
            raise TypeError('class %s is not of type Data()' % self.__class__)    
        
        # Remove the last two items, dict and object. Since the line before ensures this is a data objectt this won't break
        klasses = klasses[:-2]

        return klasses    
    
    def append(self,value,key=None):
        """ Adds new values to the classes. Can also change an already appended key
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            value
            key
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        if key is None: key = value.tag
        key = key.translate(t_table)
        if key in self: raise KeyError('key "%s" already exists' % key)
        self[key] = value        
    
    def deep_set(self,keys,val):
        """ Regresses through a list of keys the same value in various places in a dictionary.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            keys  - The keys to iterate over
            val   - The value to be set
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """               
        
        if isinstance(keys,str):
            keys = keys.split('.')
        
        data = self
         
        if len(keys) > 1:
            for k in keys[:-1]:
                data = data[k]
        
        if keys[-1][-1] ==']':
            splitkey = keys[-1].split('[')
            thing = data[splitkey[0]]
            for ii in range(1,len(splitkey)-1):
                index    = int(splitkey[ii][:-1])
                thing = thing[index]
            index    = int(splitkey[-1][:-1])
            thing[index] = val
        else:
            data[ keys[-1] ] = val
            
        return data

    def deep_get(self,keys):
        """ Regresses through a list of keys to pull a specific value out
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            keys  - The keys to iterate over
            
            Outputs:
            value - The value to be retrieved
    
            Properties Used:
            N/A    
        """          
        
        if isinstance(keys,str):
            keys = keys.split('.')
        
        data = self
         
        if len(keys) > 1:
            for k in keys[:-1]:
                data = data[k]
        
        value = data[ keys[-1] ]
        
        return value
        
    def pack_array(self,output='vector'):
        """ maps the data dict to a 1D vector or 2D column array
        
            Assumptions:
                will only pack int, float, np.array and np.matrix (max rank 2)
                if using output = 'matrix', all data values must have 
                same length (if 1D) or number of rows (if 2D), otherwise is skipped
    
            Source:
            N/A
    
            Inputs:
                output - either 'vector' (default), or 'array'
                         chooses whether to output a 1d vector or 
                         2d column array
            
            Outputs:
                array - the packed array
    
            Properties Used:
            N/A  
        
        """
        
        
        # check output type
        if not output in ('vector','array'): raise Exception('output type must be "vector" or "array"')        
        vector = output == 'vector'
        
        # list to pre-dump array elements
        M = []
        
        # valid types for output
        valid_types = ( int, float,
                        array_type,
                        matrix_type )
        
        # initialize array row size (for array output)
        size = [False]
        
        # the packing function
        def do_pack(D):
            for v in D.values(): 
                try:
                    rank = v.ndim
                except:
                    rank = 0
                    
                # type checking
                if isinstance( v, dict ): 
                    do_pack(v) # recursion!
                    continue
                elif not isinstance( v, valid_types ): continue
                elif rank > 2: continue
                # make column vectors
                v = atleast_2d_col(v)
                # handle output type
                if vector:
                    # unravel into 1d vector
                    v = v.ravel(order='F')
                else:
                    # check array size
                    size[0] = size[0] or v.shape[0] # updates size once on first array
                    if v.shape[0] != size[0]: 
                        #warn ('array size mismatch, skipping. all values in data must have same number of rows for array packing',RuntimeWarning)
                        continue
                # dump to list
                M.append(v)
            #: for each value
        
        # do the packing
        do_pack(self)
        
        # pack into final array
        if M:
            M = np.hstack(M)
        else:
            # empty result
            if vector:
                M = np.array([])
            else:
                M = np.array([[]])
        
        # done!
        return M
    
    def unpack_array(self,M):
        """ unpacks an input 1d vector or 2d column array into the data dictionary
            important that the structure of the data dictionary, and the shapes
            of the contained values are the same as the data from which the array was packed
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            M - either a 1D vector or 2D column array
            
            Outputs:
            a reference to self, updates self in place
    
            Properties Used:
            N/A    
        """           

        
        # dont require dict to have numpy
        import numpy as np
        from .Arrays import atleast_2d_col, array_type, matrix_type
        
        # check input type
        vector = M.ndim  == 1
        
        # valid types for output
        valid_types = ( int, float,
                        array_type,
                        matrix_type )
        
        # counter for unpacking
        _index = [0]
        
        # the unpacking function
        def do_unpack(D):
            for k,v in D.items():
                try:
                    rank = v.ndim
                except:
                    rank = 0
                # type checking
                if isinstance(v, dict): 
                    do_unpack(v) # recursion!
                    continue
                elif not isinstance(v,valid_types): continue
                
                # get unpack index
                index = _index[0]                
                
                # skip if too big
                if rank > 2: 
                    continue
                
                # scalars
                elif rank == 0:
                    if vector:
                        D[k] = M[index]
                        index += 1
                    else:#array
                        continue
                        #raise RuntimeError , 'array size mismatch, all values in data must have same number of rows for array unpacking'
                    
                # 1d vectors
                elif rank == 1:
                    n = len(v)
                    if vector:
                        D[k][:] = M[index:(index+n)]
                        index += n
                    else:#array
                        D[k][:] = M[:,index]
                        index += 1
                    
                # 2d arrays
                elif rank == 2:
                    n,m = v.shape
                    if vector:
                        D[k][:,:] = np.reshape( M[index:(index+(n*m))] ,[n,m], order='F')
                        index += n*m 
                    else:#array
                        D[k][:,:] = M[:,index:(index+m)]
                        index += m
                
                #: switch rank
                
                _index[0] = index

            #: for each itme
        
        # do the unpack
        do_unpack(self)
         
        # check
        if not M.shape[-1] == _index[0]: warn('did not unpack all values',RuntimeWarning)
         
        # done!
        return self     
    
    def do_recursive(self,method,other=None,default=None):
        """ Recursively applies a method of the class.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            method  - name of the method to access
    
            Outputs:
            Result  - the results of the method
    
            Properties Used:
            N/A    
        """          
    
        # result data structure
        klass = self.__class__
        if isinstance(klass,Data):
            klass = Data
        result = klass()
    
        # the update function
        def do_operation(A,B,C):
            for k,a in A.items():
                if isinstance(B,Data):
                    if k in B:
                        b = B[k]
                    else: 
                        C[k] = a
                        continue
                else:
                    b = B
                # recursion
                if isinstance(a,Data):
                    c = klass()
                    C[k] = c
                    do_operation(a,b,c)
                # method
                else:
                    if b is None:
                        c = method(a)
                    else:
                        c = method(a,b)
                    if not c is None:
                        C[k] = c
                #: if type
            #: for each key,value
    
        # do the update!
        do_operation(self,other,result)    
    
        return result


# RCAIDE/Library/Compoments/Wings/Horizontal_Tail.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Wing import Wing 
from RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_wing_moment_of_inertia import  compute_wing_moment_of_inertia

# ---------------------------------------------------------------------------------------------------------------------- 
#  Horizontal_Tail
# ----------------------------------------------------------------------------------------------------------------------    
class Horizontal_Tail(Wing):
    """
    A class representing a horizontal stabilizer surface for aircraft pitch control 
    and longitudinal stability.

    Attributes
    ----------
    tag : str
        Unique identifier for the horizontal tail, defaults to 'horizontal_tail'

    Notes
    -----
    The horizontal tail provides pitch stability and control. It inherits all geometric 
    and aerodynamic functionality from the Wing class and adds specific attributes for 
    horizontal tail operation. 

    See Also
    --------
    RCAIDE.Library.Components.Wings.Wing
        Base wing class providing core functionality
    RCAIDE.Library.Components.Wings.Stabilator
        All-moving horizontal tail variant
    RCAIDE.Library.Components.Wings.Control_Surfaces.Elevator
        Control surface typically mounted on horizontal tail
    """ 

    def __defaults__(self):
        """
        Sets default values for the horizontal tail attributes.
        """
        self.tag = 'horizontal_tail'
    
    def moment_of_inertia(wing, center_of_gravity):
        """
        Computes the moment of inertia tensor for the horizontal tail.

        Parameters
        ----------
        wing : Component
            Wing component data
        center_of_gravity : list
            Reference point coordinates for moment calculation

        Returns
        -------
        ndarray
            3x3 moment of inertia tensor
        """
        I = compute_wing_moment_of_inertia(wing, center_of_gravity) 
        return I 
    


# RCAIDE/Library/Methods/Propulsors/Converters/Rotor/compute_rotor_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports 
from RCAIDE.Framework.Core                              import Data , Units, orientation_product, orientation_transpose  
from RCAIDE.Library.Methods.Aerodynamics.Common.Lift    import compute_airfoil_aerodynamics,compute_inflow_and_tip_loss  

# package imports
import  numpy as  np 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Generalized Rotor Class
# ----------------------------------------------------------------------------------------------------------------------  
def compute_rotor_performance(propulsor,state,center_of_gravity= [[0.0, 0.0,0.0]]):
    """Analyzes a general rotor given geometry and operating conditions.

    Assumptions:
    per source

    Source:
    Drela, M. "Qprop Formulation", MIT AeroAstro, June 2006
    http://web.mit.edu/drela/Public/web/qprop/qprop_theory.pdf

    Leishman, Gordon J. Principles of helicopter aerodynamics
    Cambridge university press, 2006.

    Inputs:
    rotor.inputs.omega                    [radian/s]
    conditions.freestream.
      density                            [kg/m^3]
      dynamic_viscosity                  [kg/(m-s)]
      speed_of_sound                     [m/s]
      temperature                        [K]
    conditions.frames.
      body.transform_to_inertial         (rotation matrix)
      inertial.velocity_vector           [m/s]
    conditions.energy.
      throttle                           [-]

    Outputs:
    conditions.energy.outputs.
       number_radial_stations            [-]
       number_azimuthal_stations         [-]
       disc_radial_distribution          [m]
       speed_of_sound                    [m/s]
       density                           [kg/m-3]
       velocity                          [m/s]
       disc_tangential_induced_velocity  [m/s]
       disc_axial_induced_velocity       [m/s]
       disc_tangential_velocity          [m/s]
       disc_axial_velocity               [m/s]
       drag_coefficient                  [-]
       lift_coefficient                  [-]
       omega                             [rad/s]
       disc_circulation                  [-]
       blade_dQ_dR                       [N/m]
       blade_dT_dr                       [N]
       blade_thrust_distribution         [N]
       disc_thrust_distribution          [N]
       thrust_per_blade                  [N]
       thrust_coefficient                [-]
       azimuthal_distribution            [rad]
       disc_azimuthal_distribution       [rad]
       blade_dQ_dR                       [N]
       blade_dQ_dr                       [Nm]
       blade_torque_distribution         [Nm]
       disc_torque_distribution          [Nm]
       torque_per_blade                  [Nm]
       torque_coefficient                [-]
       power                             [W]
       power_coefficient                 [-]

    Properties Used:
    rotor.
      number_of_blades                   [-]
      tip_radius                         [m]
      twist_distribution                 [radians]
      chord_distribution                 [m]
      orientation_euler_angles           [rad, rad, rad]
    """

    # Unpack rotor blade parameters and operating conditions 
    conditions            = state.conditions 
    if 'rotor' in  propulsor:
        rotor =  propulsor.rotor
    elif 'propeller' in  propulsor:
        rotor =  propulsor.propeller
     
    propulsor_conditions  = conditions.energy[propulsor.tag]
    rotor_conditions      = propulsor_conditions[rotor.tag]
    commanded_TV          = propulsor_conditions.commanded_thrust_vector_angle
    eta                   = rotor_conditions.throttle 
    omega                 = rotor_conditions.omega
    pitch_c               = rotor_conditions.pitch_command
    B                     = rotor.number_of_blades
    R                     = rotor.tip_radius
    beta_0                = rotor.twist_distribution
    c                     = rotor.chord_distribution
    sweep                 = rotor.sweep_distribution     
    r_1d                  = rotor.radius_distribution
    tc                    = rotor.thickness_to_chord
    a_loc                 = rotor.airfoil_polar_stations
    airfoils              = rotor.airfoils 
    Na                    = rotor.number_azimuthal_stations
    nonuniform_freestream = rotor.nonuniform_freestream
    use_2d_analysis       = rotor.use_2d_analysis 

    # Check for variable pitch
    if np.any(pitch_c !=0) and not rotor.variable_pitch:
        print("Warning: pitch commanded for a fixed-pitch rotor. Changing to variable pitch rotor for weights analysis.")
        rotor.variable_pitch = True

    # Unpack freestream conditions
    rho     = conditions.freestream.density[:,0,None]
    mu      = conditions.freestream.dynamic_viscosity[:,0,None]
    a       = conditions.freestream.speed_of_sound[:,0,None]
    T       = conditions.freestream.temperature[:,0,None]
    Vv      = conditions.frames.inertial.velocity_vector
    nu      = mu/rho
    rho_0   = rho
    T_0     = T

    # Number of radial stations and segment control points
    Nr       = len(c)
    ctrl_pts = len(Vv)
    
    # Helpful shorthands
    pi      = np.pi

    # Calculate total blade pitch
    total_blade_pitch = beta_0 + pitch_c

    # Velocity in the rotor frame
    T_body2inertial         = conditions.frames.body.transform_to_inertial
    T_inertial2body         = orientation_transpose(T_body2inertial)
    V_body                  = orientation_product(T_inertial2body,Vv)
    body2thrust,orientation = rotor.body_to_prop_vel(commanded_TV) 
    T_body2thrust           = orientation_transpose(np.ones_like(T_body2inertial[:])*body2thrust)
    V_thrust                = orientation_product(T_body2thrust,V_body)

    # Check and correct for hover
    V         = V_thrust[:,0,None]
    V[V==0.0] = 1E-6

    # Non-dimensional radial distribution and differential radius
    chi           = r_1d/R
    diff_r        = np.diff(r_1d)
    deltar        = np.zeros(len(r_1d))
    deltar[1:-1]  = diff_r[0:-1]/2 + diff_r[1:]/2
    deltar[0]     = diff_r[0]/2
    deltar[-1]    = diff_r[-1]/2

    # Calculating rotational parameters
    omegar   = np.outer(omega,r_1d)
    n        = omega/(2.*pi)   # Rotations per second

    # 2 dimensional radial distribution non dimensionalized
    chi_2d         = np.tile(chi[:, None],(1,Na))
    chi_2d         = np.repeat(chi_2d[None,:,:], ctrl_pts, axis=0)
    r_dim_2d       = np.tile(r_1d[:, None] ,(1,Na))
    r_dim_2d       = np.repeat(r_dim_2d[None,:,:], ctrl_pts, axis=0)
    c_2d           = np.tile(c[:, None] ,(1,Na))
    c_2d           = np.repeat(c_2d[None,:,:], ctrl_pts, axis=0)

    # Azimuthal distribution of stations (in direction of rotation)
    psi            = np.linspace(0,2*pi,Na+1)[:-1]
    psi_2d         = np.tile(np.atleast_2d(psi),(Nr,1))
    psi_2d         = np.repeat(psi_2d[None, :, :], ctrl_pts, axis=0)

    # apply blade sweep to azimuthal position
    if np.any(np.array([sweep])!=0):
        use_2d_analysis     = True
        sweep_2d            = np.repeat(sweep[:, None], (1,Na))
        sweep_offset_angles = np.tan(sweep_2d/r_dim_2d)
        psi_2d             += sweep_offset_angles

    # Starting with uniform freestream
    ua       = 0
    ut       = 0
    ur       = 0

    # Include velocities introduced by rotor incidence angles
    if (np.any(abs(V_thrust[:,1]) >1e-3) or np.any(abs(V_thrust[:,2]) >1e-3)) and use_2d_analysis:

        # y-component of freestream in the propeller cartesian plane
        Vy  = V_thrust[:,1,None,None]
        Vy  = np.repeat(Vy, Nr,axis=1)
        Vy  = np.repeat(Vy, Na,axis=2)

        # z-component of freestream in the propeller cartesian plane
        Vz  = V_thrust[:,2,None,None]
        Vz  = np.repeat(Vz, Nr,axis=1)
        Vz  = np.repeat(Vz, Na,axis=2)

        # compute resulting radial and tangential velocities in polar frame
        utz =  -Vz*np.sin(psi_2d)
        urz =   Vz*np.cos(psi_2d)
        uty =  -Vy*np.cos(psi_2d)
        ury =   Vy*np.sin(psi_2d)

        ut +=  (utz + uty)  # tangential velocity in direction of rotor rotation
        ur +=  (urz + ury)  # radial velocity (positive toward tip)
        ua +=  np.zeros_like(ut)
        

    # Include external velocities introduced by user
    if nonuniform_freestream:
        use_2d_analysis   = True

        # include additional influences specified at rotor sections, shape=(ctrl_pts,Nr,Na)
        ua += rotor.axial_velocities_2d
        ut += rotor.tangential_velocities_2d
        ur += rotor.radial_velocities_2d

    if use_2d_analysis:
        # make everything 2D with shape (ctrl_pts,Nr,Na)

        # 2-D freestream velocity and omega*r
        V_2d   = V_thrust[:,0,None,None]
        V_2d   = np.repeat(V_2d, Na,axis=2)
        V_2d   = np.repeat(V_2d, Nr,axis=1)
        omegar = (np.repeat(np.outer(omega,r_1d)[:,:,None], Na, axis=2))

        # total velocities
        Ua     = V_2d + ua

        # 2-D blade pitch and radial distributions
        if np.size(pitch_c)>1:
            # control variable is the blade pitch, repeat around azimuth
            beta = np.repeat(total_blade_pitch[:,:,None], Na, axis=2)
        else:
            beta = np.tile(total_blade_pitch[:,:,None],(ctrl_pts,1,Na ))

        r    = np.tile(r_1d[None,:,None], (ctrl_pts, 1, Na))
        c    = np.tile(c[None,:,None], (ctrl_pts, 1, Na))
        deltar = np.tile(deltar[None,:,None], (ctrl_pts, 1, Na))

        # 2-D atmospheric properties
        a   = np.tile(np.atleast_2d(a),(1,Nr))
        a   = np.repeat(a[:, :, None], Na, axis=2)
        nu  = np.tile(np.atleast_2d(nu),(1,Nr))
        nu  = np.repeat(nu[:,  :, None], Na, axis=2)
        rho = np.tile(np.atleast_2d(rho),(1,Nr))
        rho = np.repeat(rho[:,  :, None], Na, axis=2)
        T   = np.tile(np.atleast_2d(T),(1,Nr))
        T   = np.repeat(T[:, :, None], Na, axis=2)

    else:
        # total velocities
        r      = r_1d
        Ua     = np.outer((V + ua),np.ones_like(r))
        beta   = total_blade_pitch

    # Total velocities
    Ut     = omegar - ut
    U      = np.sqrt(Ua*Ua + Ut*Ut + ur*ur)

    #---------------------------------------------------------------------------
    # COMPUTE WAKE-INDUCED INFLOW VELOCITIES AND RESULTING ROTOR PERFORMANCE
    #---------------------------------------------------------------------------
    # pack inputs
    wake_inputs                       = Data()
    wake_inputs.velocity_total        = U
    wake_inputs.velocity_axial        = Ua
    wake_inputs.velocity_tangential   = Ut
    wake_inputs.ctrl_pts              = ctrl_pts
    wake_inputs.Nr                    = Nr
    wake_inputs.Na                    = Na        
    wake_inputs.use_2d_analysis       = use_2d_analysis        
    wake_inputs.twist_distribution    = beta
    wake_inputs.chord_distribution    = c
    wake_inputs.radius_distribution   = r
    wake_inputs.speed_of_sounds       = a
    wake_inputs.dynamic_viscosities   = nu

    va, vt = rotor.Wake.evaluate(rotor,wake_inputs,conditions)
    
    # compute new blade velocities
    Wa   = va + Ua
    Wt   = Ut - vt

    lamdaw, F, _ = compute_inflow_and_tip_loss(r,R,Wa,Wt,B)

    # Compute aerodynamic forces based on specified input airfoil or surrogate
    Cl, Cdval, alpha, alpha_disc,Ma,W,Re,Re_disc = compute_airfoil_aerodynamics(beta,c,r,R,B,Wa,Wt,a,nu,airfoils,a_loc,ctrl_pts,Nr,Na,tc,use_2d_analysis) 
    
    # compute HFW circulation at the blade
    Gamma = 0.5*W*c*Cl  

    #---------------------------------------------------------------------------      
    # tip loss correction for velocities, since tip loss correction is only applied to loads in prior BET iteration
    va     = F*va
    vt     = F*vt
    lamdaw = r*(va+Ua)/(R*(Ut-vt))

    # More Cd scaling from Mach from AA241ab notes for turbulent skin friction
    Tw_Tinf     = 1. + 1.78*(Ma*Ma)
    Tp_Tinf     = 1. + 0.035*(Ma*Ma) + 0.45*(Tw_Tinf-1.)
    Tp          = (Tp_Tinf)*T
    Rp_Rinf     = (Tp_Tinf**2.5)*(Tp+110.4)/(T+110.4)
    Cd          = ((1/Tp_Tinf)*(1/Rp_Rinf)**0.2)*Cdval

    epsilon                  = Cd/Cl
    epsilon[epsilon==np.inf] = 10.

    # thrust and torque and their derivatives on the blade.
    blade_T_distribution     = rho*(Gamma*(Wt-epsilon*Wa))*deltar
    blade_Q_distribution     = rho*(Gamma*(Wa+epsilon*Wt)*r)*deltar
    blade_dT_dr              = rho*(Gamma*(Wt-epsilon*Wa))
    blade_dQ_dr              = rho*(Gamma*(Wa+epsilon*Wt)*r)


    if use_2d_analysis:
        blade_T_distribution_2d = blade_T_distribution
        blade_Q_distribution_2d = blade_Q_distribution
        blade_dT_dr_2d          = blade_dT_dr
        blade_dQ_dr_2d          = blade_dQ_dr
        blade_Gamma_2d          = Gamma
        alpha_2d                = alpha

        Va_2d = Wa
        Vt_2d = Wt
        V_disc  = np.sqrt(Va_2d**2 + Vt_2d**2)
        M_disc  = Ma
        
        Va_avg = np.average(Wa, axis=2)      # averaged around the azimuth
        Vt_avg = np.average(Wt, axis=2)      # averaged around the azimuth

        Va_ind_2d  = va
        Vt_ind_2d  = vt
        Vt_ind_avg = np.average(vt, axis=2)
        Va_ind_avg = np.average(va, axis=2)

        # set 1d blade loadings to be the average:
        blade_T_distribution    = np.mean((blade_T_distribution_2d), axis = 2)
        blade_Q_distribution    = np.mean((blade_Q_distribution_2d), axis = 2)
        blade_dT_dr             = np.mean((blade_dT_dr_2d), axis = 2)
        blade_dQ_dr             = np.mean((blade_dQ_dr_2d), axis = 2)

        # compute the hub force / rotor drag distribution along the blade
        dL_2d    = 0.5*rho*c_2d*Cd*omegar**2*deltar
        dD_2d    = 0.5*rho*c_2d*Cl*omegar**2*deltar

        rotor_drag_distribution = np.mean(dL_2d*np.sin(psi_2d) + dD_2d*np.cos(psi_2d),axis=2)

    else:
        Va_2d   = np.repeat(Wa[ :, :, None], Na, axis=2)
        Vt_2d   = np.repeat(Wt[ :, :, None], Na, axis=2)

        blade_T_distribution_2d  = np.repeat(blade_T_distribution[:, :, None], Na, axis=2)
        blade_Q_distribution_2d  = np.repeat(blade_Q_distribution[:, :, None], Na, axis=2)
        blade_dT_dr_2d           = np.repeat(blade_dT_dr[:, :, None], Na, axis=2)
        blade_dQ_dr_2d           = np.repeat(blade_dQ_dr[:, :, None], Na, axis=2)
        blade_Gamma_2d           = np.repeat(Gamma[ :, :, None], Na, axis=2)
        alpha_2d                 = np.repeat(alpha[ :, :, None], Na, axis=2)
        V_disc                   = np.sqrt(Va_2d**2 + Vt_2d**2)
        M_disc                   = np.repeat(Ma[ :, :, None], Na, axis=2)
        
        Vt_avg                  = Wt
        Va_avg                  = Wa
        Vt_ind_avg              = vt
        Va_ind_avg              = va
        Va_ind_2d               = np.repeat(va[ :, :, None], Na, axis=2)
        Vt_ind_2d               = np.repeat(vt[ :, :, None], Na, axis=2)

        # compute the hub force / rotor drag distribution along the blade
        dL    = 0.5*rho*c*Cd*omegar**2*deltar
        dL_2d = np.repeat(dL[:, :, None], Na, axis=2)
        dD    = 0.5*rho*c*Cl*omegar**2*deltar
        dD_2d = np.repeat(dD[:, :, None], Na, axis=2)

        rotor_drag_distribution = np.mean(dL_2d*np.sin(psi_2d) + dD_2d*np.cos(psi_2d),axis=2)

    # forces
    thrust                  = np.atleast_2d((B * np.sum(blade_T_distribution, axis = 1))).T
    torque                  = np.atleast_2d((B * np.sum(blade_Q_distribution, axis = 1))).T
    rotor_drag              = np.atleast_2d((B * np.sum(rotor_drag_distribution, axis=1))).T
    power                   = omega*torque

    # calculate coefficients
    D        = 2*R
    Cq       = torque/(rho_0*(n*n)*(D*D*D*D*D))
    Ct       = thrust/(rho_0*(n*n)*(D*D*D*D))
    Cp       = power/(rho_0*(n*n*n)*(D*D*D*D*D))
    Crd      = rotor_drag/(rho_0*(n*n)*(D*D*D*D))
    etap     = V*thrust/power
    A        = np.pi*(R**2 - rotor.hub_radius**2)
    FoM      = thrust*np.sqrt(thrust/(2*rho_0*A))/power  

    # prevent things from breaking
    Cq[Cq<0]                   = 0.
    Ct[Ct<0]                   = 0.
    Cp[Cp<0]                   = 0.
    thrust[omega<0.0]          = -thrust[omega<0.0]
    thrust[omega==0.0]         = 0.0
    power[omega==0.0]          = 0.0
    torque[omega==0.0]         = 0.0
    rotor_drag[omega==0.0]     = 0.0
    Ct[omega==0.0]             = 0.0
    Cp[omega==0.0]             = 0.0
    etap[omega==0.0]           = 0.
    
    
    thrust[eta[:,0]  <=0.0]    = 0.0
    power[eta[:,0]  <=0.0]     = 0.0
    torque[eta[:,0]  <=0.0]    = 0.0 
    power[eta>1.0]             = power[eta>1.0]*eta[eta>1.0]
    thrust[eta[:,0]>1.0,:]     = thrust[eta[:,0]>1.0,:]*eta[eta[:,0]>1.0,:]

    disc_loading           = thrust/(np.pi*(R**2))
    power_loading          = thrust/(power)

    # Make the thrust a 3D vector
    thrust_prop_frame      = np.zeros((ctrl_pts,3))
    thrust_prop_frame[:,0] = thrust[:,0]
    thrust_vector          = orientation_product(orientation_transpose(T_body2thrust),thrust_prop_frame)
 
    # Compute moment 
    moment_vector           = np.zeros((ctrl_pts,3))
    moment_vector[:,0]      = rotor.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = rotor.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = rotor.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, thrust_vector)
      
     
    outputs                                       = Data( 
                torque                            = torque,
                thrust                            = thrust_vector,  
                power                             = power,
                moment                            = moment,
                azimuthal_distribution            = psi, 
                rpm                               = omega /Units.rpm ,   
                tip_mach                          = omega * R / conditions.freestream.speed_of_sound, 
                efficiency                        = etap,         
                number_radial_stations            = Nr,
                orientation                       = orientation,  
                number_azimuthal_stations         = Na,
                disc_radial_distribution          = r_dim_2d,
                speed_of_sound                    = conditions.freestream.speed_of_sound,
                density                           = conditions.freestream.density,
                velocity                          = Vv,
                blade_tangential_induced_velocity = Vt_ind_avg,
                blade_axial_induced_velocity      = Va_ind_avg,
                blade_reynolds_number             = Re,
                blade_effective_angle_of_attack   = alpha,
                disc_reynolds_number              = Re_disc,
                disc_effective_angle_of_attack    = alpha_disc,
                blade_tangential_velocity         = Vt_avg,
                blade_axial_velocity              = Va_avg,
                blade_velocity                    = W,
                blade_Mach_number                 = Ma,
                disc_tangential_induced_velocity  = Vt_ind_2d,
                disc_axial_induced_velocity       = Va_ind_2d,
                disc_tangential_velocity          = Vt_2d,
                disc_axial_velocity               = Va_2d,
                disc_velocity                     = V_disc,
                disc_Mach_number                  = M_disc,
                drag_coefficient                  = Cd,
                lift_coefficient                  = Cl, 
                disc_loading                      = disc_loading, 
                power_loading                     = power_loading,      
                omega                             = omega,
                disc_circulation                  = blade_Gamma_2d,
                blade_dT_dr                       = blade_dT_dr,
                disc_dT_dr                        = blade_dT_dr_2d,
                blade_thrust_distribution         = blade_T_distribution,
                disc_thrust_distribution          = blade_T_distribution_2d, 
                thrust_per_blade                  = thrust/B,
                thrust_coefficient                = Ct,
                disc_azimuthal_distribution       = psi_2d,
                blade_dQ_dr                       = blade_dQ_dr,
                disc_dQ_dr                        = blade_dQ_dr_2d,
                blade_torque_distribution         = blade_Q_distribution,
                disc_torque_distribution          = blade_Q_distribution_2d,
                torque_per_blade                  = torque/B,
                torque_coefficient                = Cq,
                power_coefficient                 = Cp, 
                converged_inflow_ratio            = lamdaw, 
                blade_H_distribution              = rotor_drag_distribution,
                rotor_drag                        = rotor_drag,
                rotor_drag_coefficient            = Crd,
                pitch_command                     = pitch_c, 
                figure_of_merit                   = FoM, 
        ) 
    
    conditions.energy[propulsor.tag][rotor.tag] = outputs   
    
    return  



# RCAIDE/Library/Methods/Propulsors/Electric_Rotor_Propulsor/append_electric_rotor_residual_and_unknown.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
 # RCAIDE imports   
from RCAIDE.Library.Components.Propulsors.Converters.Propeller   import Propeller 
from RCAIDE.Library.Components.Propulsors.Converters.Lift_Rotor  import Lift_Rotor 
from RCAIDE.Library.Components.Propulsors.Converters.Prop_Rotor  import Prop_Rotor 

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_electric_rotor_residual_and_unknown
# ----------------------------------------------------------------------------------------------------------------------  
def append_electric_rotor_residual_and_unknown(propulsor,segment):
    '''
    
    appends the torque matching residual and unknown
    '''
    
    ones_row    = segment.state.ones_row
    rotor   = propulsor.rotor  
    if type(rotor) == Propeller:
        cp_init  = float(rotor.cruise.design_power_coefficient)
    elif (type(rotor) == Lift_Rotor) or (type(rotor) == Prop_Rotor):
        cp_init  = float(rotor.hover.design_power_coefficient)    
    segment.state.unknowns[ propulsor.tag + '_rotor_cp']                    = cp_init * ones_row(1)  
    segment.state.residuals.network[propulsor.tag +'_rotor_motor_torque'] = 0. * ones_row(1)
    
    return 


# RCAIDE/Library/Methods/Propulsors/Converters/turbine/append_turbine_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  
# ----------------------------------------------------------------------------------------------------------------------    
def append_turbine_conditions(turbine,segment,propulsor_conditions): 
    ones_row    = segment.state.ones_row 
    propulsor_conditions[turbine.tag]                                       = Conditions()
    propulsor_conditions[turbine.tag].inputs                                = Conditions()
    propulsor_conditions[turbine.tag].outputs                               = Conditions()
    propulsor_conditions[turbine.tag].inputs.fan                            = Conditions()
    propulsor_conditions[turbine.tag].inputs.fan.work_done                  = 0*ones_row(1)  
    propulsor_conditions[turbine.tag].inputs.shaft_power_off_take           = Conditions()
    propulsor_conditions[turbine.tag].inputs.shaft_power_off_take.work_done = 0*ones_row(1) 
    return 


# RCAIDE/Library/Plots/Geometry/plot_3d_rotor.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Data
from RCAIDE.Library.Plots.Geometry.Common.contour_surface_slice import contour_surface_slice
from RCAIDE.Library.Methods.Geometry.Airfoil import import_airfoil_geometry
from RCAIDE.Library.Methods.Geometry.Airfoil import compute_naca_4series

# python imports 
import numpy as np 
import plotly.graph_objects as go     

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------    
def plot_3d_rotor(rotor,save_filename = "Rotor", save_figure = False, plot_data = None,
                  show_figure  = True, plot_axis = False, cpt=0, number_of_airfoil_points = 21,
                    color_map='turbid',alpha=1):
    """ This plots a 3D surface of the  rotor

    Assumptions:
    None

    Source:
    None

    Inputs:
    axes                       - plotting axes
    rotor                      - RCAIDE rotor for which to plot the geometry
    cpt                        - control point at which to plot the rotor
    number_of_airfoil_points   - discretization of airfoil geometry 
    

    Properties Used:
    N/A
    """
    plot_propeller_only = False
    if plot_data == None: 
        print("\nPlotting rotor") 
    
        plot_propeller_only = True         
        camera        = dict(up=dict(x=0.5, y=0.5, z=1), center=dict(x=0, y=0, z=-0.5), eye=dict(x=1.5, y=1.5, z=.8))
        plot_data     = []
        
    num_B     = rotor.number_of_blades 
    af_pts    = number_of_airfoil_points-1
    dim       = len(rotor.radius_distribution)

    for i in range(num_B):
        G = generate_3d_blade_points(rotor,number_of_airfoil_points,dim,i)
        # ------------------------------------------------------------------------
        # Plot Rotor Blade
        # ------------------------------------------------------------------------
        for sec in range(dim-1):
            for loc in range(af_pts):
                X = np.array([[G.XA1[cpt,sec,loc],G.XA2[cpt,sec,loc]],
                     [G.XB1[cpt,sec,loc],G.XB2[cpt,sec,loc]]])
                Y = np.array([[G.YA1[cpt,sec,loc],G.YA2[cpt,sec,loc]],
                     [G.YB1[cpt,sec,loc],G.YB2[cpt,sec,loc]]])
                Z = np.array([[G.ZA1[cpt,sec,loc],G.ZA2[cpt,sec,loc]],
                     [G.ZB1[cpt,sec,loc],G.ZB2[cpt,sec,loc]]]) 
                 
                values      = np.ones_like(X) 
                verts       = contour_surface_slice(X, Y, Z ,values,color_map)
                plot_data.append(verts)      
            
    axis_limits = np.maximum(np.max(G.XA1), np.maximum(np.max(G.YA1),np.max(G.ZA1)))*2 
    if plot_propeller_only:
        fig = go.Figure(data=plot_data)
        fig.update_scenes(aspectmode   = 'auto',
                          xaxis_visible=plot_axis,
                          yaxis_visible=plot_axis,
                          zaxis_visible=plot_axis
                          )
        fig.update_layout( 
                 width     = 1500,
                 height    = 1500, 
                 scene = dict(
                            xaxis = dict(backgroundcolor="lightgrey", gridcolor="white", showbackground=plot_axis,
                                         zerolinecolor="white", range=[-axis_limits,axis_limits]),
                            yaxis = dict(backgroundcolor="lightgrey", gridcolor="white", showbackground=plot_axis, 
                                         zerolinecolor="white", range=[-axis_limits,axis_limits]),
                            zaxis = dict(backgroundcolor="lightgrey",gridcolor="white",showbackground=plot_axis,
                                         zerolinecolor="white", range=[-axis_limits,axis_limits])),             
                 scene_camera=camera) 
        fig.update_coloraxes(showscale=False)
        fig.update_traces(opacity = alpha) 
        if save_figure: 
            fig.write_image(save_filename + ".png")
        if show_figure:
            fig.write_html( save_filename + '.html', auto_open=True)
        return 
    else: 
        return plot_data
 
def generate_3d_blade_points(rotor,n_points,dim,i,aircraftRefFrame=True):
    """ This generates the coordinates of the blade surface for plotting in the aircraft frame (x-back, z-up)

    Assumptions:
    None

    Source:
    None

    Inputs:
    rotor            - RCAIDE rotor
    n_points         - number of points around airfoils of each blade section
    dim              - number for radial dimension
    i                - blade number
    aircraftRefFrame - boolean to convert the coordinates from rotor frame to aircraft frame 

    Properties Used:
    N/A
    """    
    # unpack 
    num_B        = rotor.number_of_blades
    airfoils     = rotor.airfoils 
    beta         = rotor.twist_distribution 
    a_o          = rotor.start_angle
    b            = rotor.chord_distribution
    r            = rotor.radius_distribution
    MCA          = rotor.mid_chord_alignment
    t            = rotor.max_thickness_distribution
    a_loc        = rotor.airfoil_polar_stations
    origin       = rotor.origin

    if rotor.clockwise_rotation:
        # negative chord and twist to give opposite rotation direction
        b    = -b    
        beta = -beta

    theta  = np.linspace(0,2*np.pi,num_B+1)[:-1] 
    flip_2 =  (np.pi/2)

    MCA_2d             = np.repeat(np.atleast_2d(MCA).T,n_points,axis=1)
    b_2d               = np.repeat(np.atleast_2d(b).T  ,n_points,axis=1)
    t_2d               = np.repeat(np.atleast_2d(t).T  ,n_points,axis=1)
    r_2d               = np.repeat(np.atleast_2d(r).T  ,n_points,axis=1)
    airfoil_le_offset  = np.repeat(b[:,None], n_points, axis=1)/2  

    # get airfoil coordinate geometry
    a_loc =  np.array(a_loc)
    if len(airfoils.keys())>0:
        xpts  = np.zeros((dim,n_points))
        zpts  = np.zeros((dim,n_points))
        max_t = np.zeros(dim)
        for af_idx,airfoil in enumerate(airfoils):
            geometry     = import_airfoil_geometry(airfoil.coordinate_file,n_points)
            locs         = np.where(a_loc == af_idx)
            xpts[locs]   = geometry.x_coordinates  
            zpts[locs]   = geometry.y_coordinates  
            max_t[locs]  = geometry.thickness_to_chord 

    else: 
        airfoil_data = compute_naca_4series('2410',n_points)
        xpts         = np.repeat(np.atleast_2d(airfoil_data.x_coordinates) ,dim,axis=0)
        zpts         = np.repeat(np.atleast_2d(airfoil_data.y_coordinates) ,dim,axis=0)
        max_t        = np.repeat(airfoil_data.thickness_to_chord,dim,axis=0)

    # store points of airfoil in similar format as Vortex Points (i.e. in vertices)
    max_t2d = np.repeat(np.atleast_2d(max_t).T ,n_points,axis=1)

    xp      = (- MCA_2d + xpts*b_2d - airfoil_le_offset)     # x-coord of airfoil
    yp      = r_2d*np.ones_like(xp)                          # radial location
    zp      = zpts*(t_2d/max_t2d)                            # former airfoil y coord
     
    commanded_thrust_vector      = np.zeros((1,1))
    rotor_vel_to_body,orientaion = rotor.prop_vel_to_body(commanded_thrust_vector)
    cpts                         = len(rotor_vel_to_body[:,0,0])

    matrix        = np.zeros((len(zp),n_points,3)) # radial location, airfoil pts (same y)
    matrix[:,:,0] = xp
    matrix[:,:,1] = yp
    matrix[:,:,2] = zp
    matrix        = np.repeat(matrix[None,:,:,:], cpts, axis=0)


    # ROTATION MATRICES FOR INNER SECTION
    # rotation about y axis to create twist and position blade upright
    trans_1        = np.zeros((dim,3,3))
    trans_1[:,0,0] = np.cos(- beta)
    trans_1[:,0,2] = -np.sin(- beta)
    trans_1[:,1,1] = 1
    trans_1[:,2,0] = np.sin(- beta)
    trans_1[:,2,2] = np.cos(- beta)
    trans_1        = np.repeat(trans_1[None,:,:,:], cpts, axis=0)

    # rotation about x axis to create azimuth locations
    trans_2 = np.array([[1 , 0 , 0],
                        [0 , np.cos(theta[i] + a_o + flip_2 ), -np.sin(theta[i] +a_o +  flip_2)],
                        [0,np.sin(theta[i] + a_o + flip_2), np.cos(theta[i] + a_o + flip_2)]])
    trans_2 = np.repeat(trans_2[None,:,:], dim, axis=0)
    trans_2 = np.repeat(trans_2[None,:,:,:], cpts, axis=0)

    # rotation about y to orient propeller/rotor to thrust angle (from propeller frame to aircraft frame)
    trans_3 =  rotor_vel_to_body
    trans_3 =  np.repeat(trans_3[:, None,:,: ],dim,axis=1) 

    trans     = np.matmul(trans_2,trans_1)
    rot_mat   = np.repeat(trans[:,:, None,:,:],n_points,axis=2)    

    # ---------------------------------------------------------------------------------------------
    # ROTATE POINTS
    if aircraftRefFrame:
        # rotate all points to the thrust angle with trans_3
        mat  =  np.matmul(np.matmul(rot_mat,matrix[...,None]).squeeze(axis=-1), trans_3)
    else:
        # use the rotor frame
        mat  =  np.matmul(rot_mat,matrix[...,None]).squeeze(axis=-1)
    # ---------------------------------------------------------------------------------------------
    # create empty data structure for storing geometry
    G = Data()

    # store node points
    G.X  = mat[:,:,:,0] + origin[0][0]
    G.Y  = mat[:,:,:,1] + origin[0][1]
    G.Z  = mat[:,:,:,2] + origin[0][2]

    G.PTS = np.zeros((cpts,len(zp),n_points,3))    
    G.PTS[:,:,:,0] =  mat[:,:,:,0] + origin[0][0]    
    G.PTS[:,:,:,1] =  mat[:,:,:,1] + origin[0][1]    
    G.PTS[:,:,:,2] =  mat[:,:,:,2] + origin[0][2]    

    # store points
    G.XA1  = mat[:,:-1,:-1,0] + origin[0][0]
    G.YA1  = mat[:,:-1,:-1,1] + origin[0][1]
    G.ZA1  = mat[:,:-1,:-1,2] + origin[0][2]
    G.XA2  = mat[:,:-1,1:,0]  + origin[0][0]
    G.YA2  = mat[:,:-1,1:,1]  + origin[0][1]
    G.ZA2  = mat[:,:-1,1:,2]  + origin[0][2]

    G.XB1  = mat[:,1:,:-1,0] + origin[0][0]
    G.YB1  = mat[:,1:,:-1,1] + origin[0][1]
    G.ZB1  = mat[:,1:,:-1,2] + origin[0][2]
    G.XB2  = mat[:,1:,1:,0]  + origin[0][0]
    G.YB2  = mat[:,1:,1:,1]  + origin[0][1]
    G.ZB2  = mat[:,1:,1:,2]  + origin[0][2]    
    
    return G
 


# RCAIDE/Library/Methods/Propulsors/Turbofan_Propulsor/append_turbofan_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_turbofan_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_turbofan_conditions(turbofan,segment):  
    ones_row    = segment.state.ones_row                  
    segment.state.conditions.energy[turbofan.tag]                               = Conditions()  
    segment.state.conditions.energy[turbofan.tag].throttle                      = 0. * ones_row(1)      
    segment.state.conditions.energy[turbofan.tag].commanded_thrust_vector_angle = 0. * ones_row(1)  
    segment.state.conditions.energy[turbofan.tag].thrust                        = 0. * ones_row(3) 
    segment.state.conditions.energy[turbofan.tag].power                         = 0. * ones_row(1) 
    segment.state.conditions.energy[turbofan.tag].moment                        = 0. * ones_row(3) 
    segment.state.conditions.energy[turbofan.tag].fuel_flow_rate                = 0. * ones_row(1)
    segment.state.conditions.energy[turbofan.tag].inputs                        = Conditions()
    segment.state.conditions.energy[turbofan.tag].outputs                       = Conditions() 
    segment.state.conditions.noise[turbofan.tag]                                = Conditions() 
    segment.state.conditions.noise[turbofan.tag].turbofan                       = Conditions() 
    segment.state.conditions.noise[turbofan.tag].turbofan.core_nozzle           = Conditions() 
    segment.state.conditions.noise[turbofan.tag].turbofan.fan_nozzle            = Conditions() 
    segment.state.conditions.noise[turbofan.tag].turbofan.fan                   = Conditions()  
    return 


# load.py 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------      

import json
import pickle
from RCAIDE.Framework.Core import Data, DataOrdered
import numpy as np
from collections import OrderedDict

# ----------------------------------------------------------------------------------------------------------------------
#  load
# ----------------------------------------------------------------------------------------------------------------------    
def load(filename,pickle_format = False):
    """Converts a JSON file into a RCAIDE data structure. 
    
        Assumptions:
            None
            
        Source:
            None
     
        Args:
            filename (string)      : file to be loaded        [unitless] 
            pickle_format (boolean): pickle file format flag  [unitless]
            
        Returns:
            data  : RCAIDE data structure [unitless]  
    """ 
    
    if pickle_format:
        load_file = filename + '.pkl' 
        with open(load_file, 'rb') as file:
            data = pickle.load(file)  
    else: 
        # Get JSON string
        f = open(filename)
        res_string = f.readline()
        f.close()    
        
        # Convert to dictionary
        res_dict = json.loads(res_string,object_pairs_hook=OrderedDict)    
        
        # Convert to RCAIDE data structure
        data = read_RCAIDE_json_dict(res_dict) 
    
    return data 

def read_RCAIDE_json_dict(res_dict):
    """Builds a RCAIDE data structure based on a dictionary from a JSON file. This is initial case.

    Assumptions:
        Dictionary was created based on a previously saved RCAIDE data structure. 
        
    Source:
        None

    Args: 
        res_dict     : Dictionary based on the RCAIDE data structure [unitless] 
        
    Returns:
        RCAIDE_data  : RCAIDE data structure [unitless]   
    """      
    keys = res_dict.keys() # keys from top level
    RCAIDE_data = Data() # initialize RCAIDE data structure
    
    # Assign all values
    for k in keys:
        k = str(k)
        v = res_dict[k]
        RCAIDE_data[k] = build_data_r(v) # recursive function
    return RCAIDE_data
 
def build_data_r(v):
    """Builds a RCAIDE data structure based on a dictionary from a JSON file. This is recursive step.

    Assumptions:
        Dictionary was created based on a previously saved RCAIDE data structure. 

    Source:
        None

    Args: 
        v     : generic value [unitless]  
        
    Returns:
        ret  :  value converted to needed format [unitless]   
    """          
    tv = type(v) # Get value type
    
    # Transform to RCAIDE data structure with appropriate types
    if tv == OrderedDict:
        keys = v.keys()
        # Recursively assign values
        ret = DataOrdered()
        for k in keys:
            k = str(k)
            ret[k] = build_data_r(v[k])
    elif tv == list:
        ret = np.array(v)
    elif (tv == str): 
        ret = str(v)
    elif (tv == bool):
        ret = v
    elif tv == type(None):
        ret = None
    elif (tv == float) or (tv == int):
        ret = v        
    else:
        raise TypeError('Data type not expected in RCAIDE JSON structure')

    return ret


# RCAIDE/Library/Compoments/Thermal_Management/Accessories/Pump.py
# 
# Created: March 2024  S. Shekar 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
from RCAIDE.Framework.Core import Data

# ----------------------------------------------------------------------
#  Pump
# ----------------------------------------------------------------------
class Pump(Data):
    """
    A class representing a coolant pump for thermal management systems.

    Attributes
    ----------
    tag : str
        Unique identifier for the pump component, defaults to 'Pump'
        
    efficiency : float
        Overall pump efficiency including mechanical and electrical losses, 
        defaults to 1.0

    Notes
    -----
    The pump class models liquid circulation devices used in thermal management 
    systems. It provides functionality for:
    
    * Power consumption calculation
    * Performance modeling based on operating conditions
    * Integration with liquid cooling systems

    **Definitions**

    'Pressure Differential'
        The difference in pressure across the pump
        
    'Mass Flow Rate'
        Rate of coolant mass flow through the pump

    See Also
    --------
    RCAIDE.Library.Components.Thermal_Management.Accessories.Fan
        Similar component for gas (air) cooling systems
    """

    def __defaults__(self):
        """
        Sets default values for the pump attributes.
        """
        self.tag        = 'Pump'
        self.efficiency = 1.0
        return
   
    def compute_power_consumed(pressure_differential, density, mass_flow_rate, efficiency):
        """
        Calculates the power consumed by the pump.

        Parameters
        ----------
        pressure_differential : float
            Pressure rise across the pump
            
        density : float
            Coolant density
            
        mass_flow_rate : float
            Mass flow rate through the pump
            
        efficiency : float
            Overall pump efficiency

        Returns
        -------
        float
            Power consumed by the pump

        Notes
        -----
        Uses the standard pump power equation:
        Power = (mass_flow_rate * pressure_differential) / (density * efficiency)
        """
        return mass_flow_rate * pressure_differential / (density * efficiency)


# RCAIDE/Library/Components/Propulsors/Converters/Combustor.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from RCAIDE.Library.Components                      import Component
from RCAIDE.Library.Methods.Propulsors.Converters.Combustor.append_combustor_conditions import  append_combustor_conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  Combustor
# ---------------------------------------------------------------------------------------------------------------------- 
class Combustor(Component):
    """
    A combustor component model for gas turbine engines that simulates the combustion process.

    Attributes
    ----------
    tag : str
        Identifier for the combustor. Default is 'Combustor'.
        
    alphac : float
        Combustor entrance angle [rad]. Default is 0.0.
        
    turbine_inlet_temperature : float
        Temperature at turbine inlet [K]. Default is 1500.
        
    area_ratio : float
        Ratio of combustor exit to inlet area. Default is 1.0.
        
    axial_fuel_velocity_ratio : float
        Ratio of axial fuel velocity to inlet velocity. Default is 0.0.
        
    fuel_velocity_ratio : float
        Ratio of fuel velocity to inlet velocity. Default is 0.0.
        
    burner_drag_coefficient : float
        Drag coefficient of the burner. Default is 0.0.
        
    absolute_sensible_enthalpy : float
        Absolute sensible enthalpy [J/kg]. Default is 0.0.
        
    diameter : float
        Combustor diameter [m]. Default is 0.2.
        
    length : float
        Combustor length [m]. Default is 0.3.
        
    fuel_equivalency_ratio : float
        Fuel-to-air equivalency ratio. Default is 0.3.
        
    number_of_combustors : int
        Number of combustor cans. Default is 30.
        
    f_air_PZ : float
        Fraction of total air entering Primary Zone. Default is 0.18.
        
    FAR_st : float
        Stoichiometric Fuel to Air ratio. Default is 0.068.
        
    N_comb : int
        Number of can-annular combustors. Default is 10.
        
    N_PZ : int
        Number of PSR in the Primary Zone. Default is 8.
        
    A_PZ : float
        Primary Zone cross-sectional area [m²]. Default is 0.15.
        
    L_PZ : float
        Primary Zone length [m]. Default is 0.0153.
        
    N_SZ : int
        Number of dilution air inlets in the Secondary Zone. Default is 3.
        
    A_SZ : float
        Secondary Zone cross-sectional area [m²]. Default is 0.15.
        
    L_SZ : float
        Secondary Zone length [m]. Default is 0.075.
        
    phi_SZ : float
        Equivalence Ratio in the Secondary Zone. Default is 0.2.
        
    S_PZ : float
        Mixing parameter in the Primary Zone. Default is 0.6.
        
    F_SC : float
        Fuel scaler. Default is 0.425.
        
    number_of_assigned_PSR_1st_mixers : int
        Number of assigned PSRs to first row mixers. Default is 2.
        
    number_of_assigned_PSR_2nd_mixers : int
        Number of assigned mixers to second row mixers. Default is 2.

    Notes
    -----
    The Combustor class models the combustion process in gas turbine engines,
    splitting the combustor into primary and secondary zones. It uses a Chemical
    Reactor Network (CRN) approach with Perfectly Stirred Reactors (PSR) and Plug Flow Reactors (PFR) for
    modeling the combustion process.

    **Definitions**

    'PSR'
        Perfectly Stirred Reactor - A reactor model assuming perfect mixing
    
    'PZ'
        Primary Zone - Initial combustion region
        
    'SZ'
        Secondary Zone - Dilution region where the combustion is completed

    'FAR'
        Fuel-to-Air Ratio
    
    'PFR'
        Plug Flow Reactor - A reactor model assuming no mixing in axial direction

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reaction_Network.evaluate_cantera
    """
    
    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
        None

        Source:
        None 
        """         
        
        self.tag                               = 'Combustor' 
        self.alphac                            = 0.0
        self.turbine_inlet_temperature         = 1500
        self.area_ratio                        = 1.0
        self.axial_fuel_velocity_ratio         = 0.0
        self.fuel_velocity_ratio               = 0.0
        self.burner_drag_coefficient           = 0.0
        self.absolute_sensible_enthalpy        = 0.0 
        self.diameter                          = 0.2
        self.length                            = 0.3
        self.fuel_equivalency_ratio            = 0.3 
        self.number_of_combustors              = 30 
                                               
        self.f_air_PZ                          = 0.18                                                  # [-]       Fraction of total air present in the combustor that enters the Primary Zone         
        self.FAR_st                            = 0.068                                                 # [-]       Stoichiometric Fuel to Air ratio
        self.N_comb                            = 10                                                    # [-]       Number of can-annular combustors
        self.N_PZ                              = 8                                                     # [-]       Number of PSR (EVEN, must match the number of PSR below)
        self.A_PZ                              = 0.15                                                  # [m**2]    Primary Zone cross-sectional area     
        self.L_PZ                              = 0.0153                                                # [m]       Primary Zone length  
        self.N_SZ                              = 3                                                     # [-]       Number of dilution air inlets        
        self.A_SZ                              = 0.15                                                  # [m**2]    Secondary Zone cross-sectional area
        self.L_SZ                              = 0.075                                                 # [m]       Secondary Zone length  
        self.phi_SZ                            = 0.2                                                   # [-]       Equivalence Ratio for PFR    phi_PZ_des              = 0.6                                                   # [-]       Primary Zone Design Equivalence Ratio
        self.S_PZ                              = 0.6                                                   # [-]       Mixing parameter, used to define the Equivalence Ratio standard deviation  
        self.F_SC                              = 0.425                                                 # [-]       Fuel scaler
        self.number_of_assigned_PSR_1st_mixers = 2                                                     # [-]       Number of assigned PSRs to each mixer in the first row of mixers (CRN network model)
        self.number_of_assigned_PSR_2nd_mixers = 2                                                     # [-]       Number of assigned mixers to each mixer in the second row of mixers (CRN network model)
    
    def append_operating_conditions(self,segment,propulsor):
        """
        Appends operating conditions to the combustor.
        """
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_combustor_conditions(self,segment,propulsor_conditions)
        return


# RCAIDE/Library/Attributes/Propellants/Alcohol_Mixture.py
#  
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 
import RCAIDE
from .Propellant import Propellant 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Propanol Propellant Class
# ----------------------------------------------------------------------------------------------------------------------   
class Alcohol_Mixture(Propellant):
    """Alcohol_Mixture class propellant  
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            Density at 20C 1 atm
        
        Source: 
    
        """    
        self.tag                        = 'Alcohol_Mixture'
        self.reactant                   = 'O2'
        self.propellant_1               = RCAIDE.Library.Attributes.Propellants.Ethanol() 
        self.propellant_1_mass_fraction = 0.5
        self.propellant_2               = RCAIDE.Library.Attributes.Propellants.Propanol()
        self.propellant_2_mass_fraction = 0.5
        self.density                    = self.compute_mixture_density()                                 # kg/m^3 (15 C, 1 atm)
        self.specific_energy            = self.compute_mixture_specific_energy()                         # J/kg
        self.energy_density             = self.compute_mixture_energy_density()                          # J/m^3
        self.lower_heating_value        = self.compute_mixture_lower_heating_value()                     # J/kg
        
    def compute_mixture_density(self):
        p1    = self.propellant_1 
        p2    = self.propellant_2
        rho1  = p1.density
        rho2  = p2.density
        X1    = self.propellant_1_mass_fraction
        X2    = self.propellant_2_mass_fraction

        rho_mix = ((X1/rho1) + (X2/rho2) ) ** (-1)
        return rho_mix 

    def compute_mixture_specific_energy(self):
        p1    = self.propellant_1 
        p2    = self.propellant_2
        e1    = p1.specific_energy
        e2    = p2.specific_energy
        X1    = self.propellant_1_mass_fraction
        X2    = self.propellant_2_mass_fraction

        e_mix =  X1 * e1 +  X2 * e2

        return e_mix

    def compute_mixture_energy_density(self): 
        e_mix   =  self.compute_mixture_specific_energy()
        rho_mix =  self.compute_mixture_density()

        U_mix = e_mix * rho_mix 
        return U_mix

    def compute_mixture_lower_heating_value(self):
        p1    = self.propellant_1 
        p2    = self.propellant_2
        LHV1  = p1.lower_heating_value
        LHV2  = p2.lower_heating_value
        X1    = self.propellant_1_mass_fraction
        X2    = self.propellant_2_mass_fraction

        LHV_mix =  X1 * LHV1 +  X2 * LHV2
        return LHV_mix
    
    def compute_all(self):
        self.density             = self.compute_mixture_density()                                 # kg/m^3 (15 C, 1 atm)
        self.specific_energy     = self.compute_mixture_specific_energy()                         # J/kg
        self.energy_density      = self.compute_mixture_energy_density()                          # J/m^3
        self.lower_heating_value = self.compute_mixture_lower_heating_value()                     # J/kg



# RCAIDE/Library/Plots/Noise/plot_2D_noise_contour.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units  

# python imports 
import numpy as np   
import matplotlib.pyplot as plt  
import matplotlib.colors
import matplotlib.colors as colors  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_2D_noise_contour(noise_data,
                       noise_level              = None ,
                       min_noise_level          = 35,  
                       max_noise_level          = 90, 
                       noise_scale_label        = None,
                       save_figure              = False,
                       show_figure              = True,
                       save_filename            = "2D_Noise_Contour",
                       show_elevation           = False,
                       use_lat_long_coordinates = True,  
                       colormap                 = 'jet',
                       file_type                = ".png",
                       width                    = 10, 
                       height                   = 7,
                       *args, **kwargs): 
    """This plots a 2D noise contour of a noise level 

    Assumptions:
    None

    Source:
    None

    Inputs: 
       noise_data        - noise data structure 
       noise_level       - noise level (dBA, DNL, SENEL etc)
       min_noise_level   - minimal noise level 
       max_noise_level   - maximum noise level 
       noise_scale_label - noise level label 
       save_figure       - save figure 
       show_figure       - show figure 
       save_filename     - save file flag
       show_trajectory   - plot aircraft trajectory flag
       show_microphones  - show microhpone flag 

    Outputs:
       Plots

    Properties Used:
    N/A
    """      
    
    elevation       = noise_data.microphone_locations[:,:,2]/Units.ft      
    colors_undersea = plt.cm.terrain(np.linspace(0, 0.17, 56))
    colors_land     = plt.cm.terrain(np.linspace(0.25, 1, 200))  
    colors          = np.vstack((colors_undersea, colors_land))
    cut_terrain_map = matplotlib.colors.LinearSegmentedColormap.from_list('cut_terrain', colors) 
    norm = FixPointNormalize(sealevel=0,vmax=np.max(elevation),vmin=np.min(elevation))  
     
    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    
    axis = fig.add_subplot(1,1,1) 
    
    noise_levels   = np.linspace(min_noise_level,max_noise_level,10)  
    noise_cmap     = plt.get_cmap('turbo')
    noise_new_cmap = truncate_colormap(noise_cmap,0.0, 1.0) 
     
    if use_lat_long_coordinates and (noise_data.topography_file != None ):
        LAT  = noise_data.microphone_coordinates[:,:,0]
        LONG = noise_data.microphone_coordinates[:,:,1]
        axis.set_xlabel('Longitude [°]')
        axis.set_ylabel('Latitude [°]') 
    else:
        LAT  = noise_data.microphone_locations[:,:,0]
        LONG = noise_data.microphone_locations[:,:,1]
        axis.set_xlabel('x [m]')
        axis.set_ylabel('y [m]')  
    
    if show_elevation:
        CS_1  = axis.contourf(LONG,LAT,elevation,cmap =cut_terrain_map,norm=norm,levels = 20, alpha=0.5)  
        cbar = fig.colorbar(CS_1, ax=axis)     
        cbar.ax.set_ylabel('Elevation above sea level [ft]', rotation =  90)  

    # plot aircraft noise levels   
    CS_2    = axis.contourf(LONG,LAT,noise_level ,noise_levels,cmap = noise_new_cmap)     
    cbar    = fig.colorbar(CS_2, ax=axis)        
    cbar.ax.set_ylabel(noise_scale_label, rotation =  90) 
        
    fig.tight_layout()  
    if save_figure: 
        figure_title  = save_filename
        plt.savefig(figure_title + file_type )     
         
    return fig       

# ------------------------------------------------------------------ 
# Truncate colormaps
# ------------------------------------------------------------------  
def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    return new_cmap


class FixPointNormalize(matplotlib.colors.Normalize):
    """ 
    Inspired by https://stackoverflow.com/questions/20144529/shifted-colorbar-matplotlib
    Subclassing Normalize to obtain a colormap with a fixpoint 
    somewhere in the middle of the colormap.
    This may be useful for a `terrain` map, to set the "sea level" 
    to a color in the blue/turquise range. 
    """
    def __init__(self, vmin=None, vmax=None, sealevel=0, col_val = 0.21875, clip=False):
        # sealevel is the fix point of the colormap (in data units)
        self.sealevel = sealevel
        # col_val is the color value in the range [0,1] that should represent the sealevel.
        self.col_val = col_val
        matplotlib.colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        x, y = [self.vmin, self.sealevel, self.vmax], [0, self.col_val, 1]
        return np.ma.masked_array(np.interp(value, x, y))   

def colorax(vmin, vmax):
    return dict(cmin=vmin,cmax=vmax)


# RCAIDE/Library/Methods/Propulsors/Turboprop_Propulsor/append_turboprop_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_turboprop_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_turboprop_conditions(turboprop,segment):  
    ones_row    = segment.state.ones_row                  
    segment.state.conditions.energy[turboprop.tag]                               = Conditions()  
    segment.state.conditions.energy[turboprop.tag].throttle                      = 0. * ones_row(1)     
    segment.state.conditions.energy[turboprop.tag].commanded_thrust_vector_angle = 0. * ones_row(1)   
    segment.state.conditions.energy[turboprop.tag].power                         = 0. * ones_row(1) 
    segment.state.conditions.energy[turboprop.tag].fuel_flow_rate                = 0. * ones_row(1)
    segment.state.conditions.energy[turboprop.tag].inputs                        = Conditions()
    segment.state.conditions.energy[turboprop.tag].outputs                       = Conditions() 
    segment.state.conditions.noise[turboprop.tag]                                = Conditions() 
    segment.state.conditions.noise[turboprop.tag].turboprop                      = Conditions() 
    segment.state.conditions.noise[turboprop.tag].turboprop.core_nozzle          = Conditions()   
    return 